//
// (C) 2007 Mike Brent aka Tursi aka HarmlessLion.com
// This software is provided AS-IS. No warranty
// express or implied is provided.
//
// This notice defines the entire license for this code.
// All rights not explicity granted here are reserved by the
// author.
//
// You may redistribute this software provided the original
// archive is UNCHANGED and a link back to my web page,
// http://harmlesslion.com, is provided as the author's site.
// It is acceptable to link directly to a subpage at harmlesslion.com
// provided that page offers a URL for that purpose
//
// Source code, if available, is provided for educational purposes
// only. You are welcome to read it, learn from it, mock
// it, and hack it up - for your own use only.
//
// Please contact me before distributing derived works or
// ports so that we may work out terms. I don't mind people
// using my code but it's been outright stolen before. In all
// cases the code must maintain credit to the original author(s).
//
// -COMMERCIAL USE- Contact me first. I didn't make
// any money off it - why should you? ;) If you just learned
// something from this, then go ahead. If you just pinched
// a routine or two, let me know, I'll probably just ask
// for credit. If you want to derive a commercial tool
// or use large portions, we need to talk. ;)
//
// If this, itself, is a derived work from someone else's code,
// then their original copyrights and licenses are left intact
// and in full force.
//
// http://harmlesslion.com - visit the web page for contact info
//
//*****************************************************
//* Classic 99 - TI Emulator for Win32				  *
//* by M.Brent                                        *
//*                                                   *
//* Version QI3.11 - 19 Nov 07			              *
//*                                                   *
//* Thanks to many people - see documentation         *
//*****************************************************

// Win32 Code for MSVC 6.0

// Scratchpad RAM is now at 0x8300
// any patches that want to access it directly (not through ROMWORD or RCPUBYTE)
// must take note of this or they will fail

#pragma warning (disable: 4113 4761 4101)

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0400

// This enables the old floating point speedups and console interrupt simulator
//#define ALLOW_OLD_HACKS

////////////////////////////////////////////
// Includes
////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <mmsystem.h>
#include <process.h>
#include <malloc.h>
#include <dsound.h>
#include <time.h>
#include <math.h>

#include "resource.h"
#include "tiemul.h"
#include "tms5220\5220intf.h"
#include "rs232_pio.h"
#include "keyboard\kb.h"
#include "keyboard\ti.h"
#include "ams.h"

extern "C" {
#include "sdgrom.h"
extern unsigned char bExtended; 
};

////////////////////////////////////////////
// Globals
// These don't all NEED to be globals, but I'm only cleaning up the code, 
// not re-writing it all from scratch.
////////////////////////////////////////////

// Win32 Stuff
HINSTANCE hInstance;						// global program instance
HINSTANCE hPrevInstance;					// prev instance (always null so far)
extern BOOL CALLBACK DebugBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
extern void DebugUpdateThread(void*);

// speech 
#define SPEECHUPDATECOUNT (max_cpf/5)
INT16 SpeechTmp[SPEECHRATE*2];				// two seconds worth of buffer
int nSpeechTmpPos=0;
INT16 DACTmp[500];							// just a tiny DAC buffer to speed mixing
int nDACPos=0;
int nDACCountDown;
FILE *DacLog;
HANDLE hSpeechBufferClearEvent=INVALID_HANDLE_VALUE;		// notification of speech buffer looping

HMODULE hSpeechDll;											// Handle to speech DLL
void (*SpeechInit)(Byte *pROM, int nRomLen, int BufLen,int SampleRate);	// Pointer to SpeechInit function
void (*SpeechStop)(void);									// Pointer to SpeechStop function
Byte (*SpeechRead)(void);									// Pointer to SpeechRead function
bool (*SpeechWrite)(Byte b, bool f);						// Pointer to SpeechWrite function
void (*SpeechProcess)(Byte *pBuf, int nLen);				// Pointer to SpeechProcess function
HANDLE hWakeupEvent=NULL;									// used to sleep the CPU when not busy
volatile signed long cycles_left=0;							// runs the CPU throttle
volatile unsigned long total_cycles=0;						// used for interrupts
unsigned long speech_cycles=0;								// used to sync speech
bool total_cycles_looped=false;
bool bDebugAfterStep=false;

// debug
int BreakPointType[MAX_BREAKPOINTS];	
int BreakPoint[MAX_BREAKPOINTS];
int nBreakPoints=0;

// Must remain compatible with LARGE_INTEGER - just here
// to make QuadPart unsigned ;)
typedef union {
    struct {
        DWORD LowPart;
        LONG HighPart;
    };
    struct {
        DWORD LowPart;
        LONG HighPart;
    } u;
    unsigned __int64 QuadPart;
} MY_LARGE_INTEGER;

// Memory
Byte CPU[65536];							// Main CPU Memory
Byte CPU2[8192];							// Cartridge space bank-switched (ROM >6000 space)
Byte ROMMAP[65536];							// Write-protect map of CPU space
Byte CRU[4096];								// CRU space
Byte SPEECH[65536];							// Speech Synth ROM
Byte DSR[16][16384];						// 16 CRU bases, up to 16k each (ROM >4000 space)
int  nDSRBank[16];							// Is the DSR bank switched?
struct GROMType GROMBase[17];				// support 16 GROM bases (there is room for 256 of them!), plus 1 for PCODE
int  nSystem=1;			// Which system do we default to?
int  nCartGroup=0;		// Which cart group?
int	 nCart=-1;			// Which cart is loaded (-1=none)
struct DISKS *pMagicDisk=NULL;	// which magic disk is loaded?
bool fKeyEverPressed=false;	// used to suppress warning when changing cartridges
int  nLoadedUserCarts=0;
int CRUTimerTicks=0;		// used for 9901 timer

unsigned char DummyROM[6]={
	0x83, 0x00,			// >0000	reset vector, workspace
	0x00, 0x04,			// >0002	reset vector, address
	0x10, 0xff			// >0004	JMP @>0004
};

int KEYS[2][8][8]= {  
{
// Keyboard - 99/4 - no PS/2 emulation :)
/* unused */	VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,

/* 1 */			'N', 'H', 'U', '7', 'C', 'D', 'R', '4',
/* Joy 1 */		VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,
/* 3 */			VK_OEM_PERIOD, 'K', 'O', '9', 'Z', 'A', 'W', '2',

/* Joy 2 */		VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,
	
/* 5 */			'M', 'J', 'I', '8', 'X', 'S', 'E', '3',
/* 6 */			'B', 'G', 'Y', '6', 'V', 'F', 'T', '5',
/* 7 */			VK_RETURN, 'L', 'P', '0', VK_SHIFT, VK_SPACE, 'Q', '1'
},
{
// Keyboard - 99/4A
/* Joy 2 */		VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,

/* 1 */			'M', 'J', 'U', '7', '4', 'F', 'R', 'V',					// MJU7 4FRV
/* 2 */			VK_OEM_2, VK_OEM_1, 'P', '0', '1', 'A', 'Q', 'Z',		// /;P0 1AQZ
/* 3 */			VK_OEM_PERIOD, 'L', 'O', '9', '2', 'S', 'W', 'X',		// .LO9 2SWX

/* Joy 1 */		VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,
	
/* 5 */			VK_OEM_COMMA, 'K', 'I', '8', '3', 'D', 'E', 'C',		// ,KI8 3DEC
/* 6 */			'N', 'H', 'Y', '6', '5', 'G', 'T', 'B',					// NHY6 5GTB
/* 7 */			VK_OEM_PLUS, VK_SPACE, VK_RETURN, VK_ESCAPE, VK_MENU, VK_SHIFT, VK_CONTROL, VK_ESCAPE 
}																		// = rx fscx
};

char key[256];										// keyboard state buffer

// Win32 Joysticks
JOYINFOEX myJoy;
int fJoy;
int joy1mode, joy2mode;

// Audio
int latch_byte;										// latched byte
int volume_mult;									// Volume multiplier
int nSampleIdx[3];									// Which sample (per voice)
bool fSampleShouldLoop[3];							// And whether to loop it
bool fVoiceSimulatedDAC[3];							// true if frequency is so high we should use DAC mode
int  DAC[3];										// actual DAC level (0 (silent) - 15 (loudest) )
LPDIRECTSOUND lpds;									// DirectSound handle
LPDIRECTSOUNDBUFFER voice[3];						// Voice buffers
LPDIRECTSOUNDBUFFER noise[8];						// Noise buffers
LPDIRECTSOUNDBUFFER speechbuf;						// speech structure
// Used to halt the CPU when writing to the synth too quicky
bool CPUSpeechHalt=false;
Byte CPUSpeechHaltByte=0;

// Assorted
char qw[80];										// temp string
int quitflag;										// quit flag
char lines[34][DEBUGLEN];							// debug lines
bool bDebugDirty;									// whether debug has changed
Word disasm[20];									// last 20 addresses for disasm
volatile int bank=0;								// Cartridge bank switch
int xb = 0;											// Is second bank (XB) loaded?
bool bUseSDCard=false;								// Whether SDGROM emulation is active
int grombanking = 0;								// Did we load multiple GROM bases?
int nCurrentDSR=-1;									// Which DSR Bank are we on?
unsigned int index1;								// General counter variable
int drawspeed=0;									// flag used in display updating
int max_cpf=DEFAULT_CPF;							// Maximum cycles per frame
int oldmax=max_cpf;									// copy of same
int idling=0;										// Set if an IDLE has occured
int speedup_float = 0;								// speedup floating point operations
int simulate_int = 0;								// simulate console interrupt (faster)
int slowdown_keyboard = 1;							// slowdown keyboard autorepeat in the GROM code
int cpucount, cpuframes;							// CPU counters for timing
int timercount;										// Used to estimate runtime

int timer9901;										// 9901 interrupt timer
int starttimer9901;									// and it's set time
int timer9901IntReq;								// And whether it is requesting an interrupt
int timer9901IntOk;									// Whether it's okay to generate a timer interrupt
int keyboard=KEY_994A_PS2;							// keyboard map (0=99/4, 1=99/4A, 2=99/4A PS/2 (see enum in .h))
int ps2keyboardok=1;								// whether to allow PS2 keyboards

int sams_enabled = 1;								// memory type (0 = disabled, 1 SAMS enabled)
int sams_size = 3;									// SAMS emulation memory size (0 = 128k, 1 = 256k, 2 = 512k, 3 = 1024k)

int retrace_count=0;								// count on the 60hz timer

int PauseInactive;									// what to do when the window is inactive
int SpeechEnabled;									// whether speech is enabled
volatile int CPUThrottle;							// Whether or not the CPU is throttled
volatile int SystemThrottle;						// Whether or not the VDP is throttled

time_t STARTTIME, ENDTIME;
volatile long ticks;

ATOM myClass;										// Window Class
HWND myWnd, dbgWnd;									// Handle to windows
HDC myDC;											// Handle to Device Context
int fontX, fontY;									// Non-proportional font x and y size
int wndXtrim=0, wndYtrim=0;							// Window trim, used for sizing

char AVIFileName[256]="C:\\TI99AVI.AVI";			// AVI Filename

char *PasteString;									// Used for Edit->Paste
char *PasteIndex;
int PasteCount;

char diskpath[10][256];								// path to disk directories or images
char disktype[11];									// i=image, v=files with v9t9 header
													// t=files with TIFILES header,
													// r=raw program files without header
													// (load will attempt to autodetect)
unsigned long myThread;								// timer thread
HANDLE MyMutex;										// Synchronization mutex

#ifdef ALLOW_OLD_HACKS
// Console Interrupt Routine Simulator
int num_sprites;									// Number of active sprites
int y_speed;										// X pixel speed
int x_speed;										// Y pixel speed
int t1, t1a, t1b, t2, t2a, t2b;						// temp variables
int sprite_y, sprite_x, offset;						// Sprite location
int ty, tx;											// Temp location
int snd_address;									// Sound chip data address
Byte *sndlist;										// The list itself
Byte data;											// temp data
#endif

extern const char *pCurrentHelpMsg;
extern int VDPDebug;
extern int nCycleCount;
extern int TVScanLines;

// 22khz buffer, 1 seconds worth
#define AUDIO_BUFFER_SIZE 22050
#define INIFILE ".\\classic99.ini"

///////////////////////////////////
// Built-in Cart library
///////////////////////////////////

// ROMs to always load
struct IMG AlwaysLoad[] = {
	{	IDR_AMI99DSK,	0x1100, 0x0130,	TYPE_DSR	, 0},
//	{	IDR_RS232,		0x1300, 0x0900, TYPE_DSR	, 0},
	{	IDR_SPCHROM,	0x0000,	0x8000,	TYPE_SPEECH	, 0},
	{	IDR_PGROM,		0x0000, 0xF800, TYPE_PCODEG , 0},
};

// Extra files to support certain cartridges
// These files, when the list is loaded, can be loaded as if they were
// on the disk without the disk actually needing it
// Currently these can only be loaded program image files!
struct DISKS Disk_EA[] = {
	{	"DSK1.ASSM1",	IDR_ASSM1	},
	{	"DSK1.ASSM2",	IDR_ASSM2	},
	{	"DSK1.EDIT1",	IDR_EDIT1	},
	{	"",				0			},
};

struct DISKS Disk_SSA[] = {
	{	"DSK1.ACER_C",	IDR_ACERC	},
	{	"DSK1.ACER_P",	IDR_ACERP	},
	{	"DSK1.SSD",		IDR_SSD		},
	{	"DSK1.SSE",		IDR_SSE		},
	{	"",				0			},
};

struct DISKS Disk_Tunnels[] = {
	{	"DSK1.PENNY",	IDR_PENNY	},
	{	"DSK1.QUEST",	IDR_QUEST	},

	{	"",				0			},
};

// Actual cartridge definitions (broken into categories)
struct CARTS Users[100];		// these are loaded dynamically

struct CARTS Systems[] = {
	// very slow to start - messes with the 9901? Often fails
	{	"TI-99/4",	{	
						{	IDR_CON4R0,		0x0000, 0x2000,	TYPE_ROM	, -1},

						{	IDR_CON4G0,		0x0000, 0x2000,	TYPE_GROM	, -1},
						{	IDR_CON4G1,		0x2000,	0x2000,	TYPE_GROM	, -1},
						{	IDR_CON4G2,		0x4000,	0x2000,	TYPE_GROM	, -1},
					},
		NULL,
		NULL
	},

	{	"TI-99/4A",	{	{	IDR_994AGROM,	0x0000, 0x6000,	TYPE_GROM	, -1},
						{	IDR_994AROM,	0x0000,	0x2000,	TYPE_ROM	, -1},
						{	0,				0x0000,	0x0000, TYPE_NONE	, -1},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, -1},
					},
		NULL,
		NULL
	},

	{	"TI-99/4A V2.2",{{	IDR_CON22R0,	0x0000, 0x2000,	TYPE_ROM	, -1},
						{	IDR_CON22G0,	0x0000,	0x2000,	TYPE_GROM	, -1},
						{	IDR_CON22G1,	0x2000,	0x2000,	TYPE_GROM	, -1},
						{	IDR_CON22G2,	0x4000,	0x2000,	TYPE_GROM	, -1},
					},
		NULL,
		NULL
	},
};

struct CARTS Apps[] = {
#if 0
	// not working yet
	{	"AMS Test 2.0",{{	IDR_ROMS5,		0x6000,	0x0030,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_AMS	, 0},	// address ignored on AMS cards, 0xffff flags no load, activate card
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		"AMS Card will automatically be enabled."
	},
#endif

	{	"Demonstration",{{	IDR_DEMOG,		0x6000, 0x8000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Diagnostics",{	{	IDR_DIAGNOSG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		"The maintenance tests are intended for use with external hardware, and are not supported. They will hang the emulator. Use File->Reset to bring it back."
	},
						
	{	"Editor/Assembler",{{IDR_TIEAG,		0x6000,	0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		Disk_EA,
		"The Editor and Assembler files are built-in. It is not currently possible to use the assembler due to incomplete disk emulation."
	},

	{	"Extended BASIC",{{	IDR_TIEXTG,		0x6000,	0x8000,	TYPE_GROM	, 0},
						{	IDR_TIEXTC,		0x6000,	0x2000,	TYPE_ROM	, 0},
						{	IDR_TIEXTD,		0x6000,	0x2000,	TYPE_XB		, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Home Finance",{{	IDR_HOMEG,		0x6000, 0x4000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Mini Memory",{	{	IDR_MINIMEMG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	IDR_MINIMEMC,	0x6000,	0x1000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		"The Line-By-Line assembler is not yet available."
	},

	{	"P-Code Card",{	{	IDR_PCODEC,		0x1F00,	0x2000,	TYPE_DSR	, 0},
						{	IDR_PCODED,		0x1F00, 0x2000,	TYPE_DSR2	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		"This device does not work at all yet. Select another cartridge to restore the system."
	},

	{	"Terminal Emulator 2",{{IDR_TE2G,	0x6000, 0xA000,	TYPE_GROM	, 0},
						{	IDR_TE2C,		0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		"Serial is not supported yet."
	},

	{	"TI Logo ][",{	{	IDR_LOGOG,		0x6000, 0x6000,	TYPE_GROM	, 0},
						{	IDR_LOGOC,		0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

};

struct CARTS Games[] = {
	{	"Alpiner",	{	{	IDR_ALPINERG,	0x6000, 0x8000,	TYPE_GROM	, 0},
						{	IDR_ALPINERC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},
	
	{	"A-Maze-Ing",{	{	IDR_AMAZEG,		0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"BlackJack&&Poker",{{IDR_BLACKJACK,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Car Wars",	{	{	IDR_CARWARS,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Chisholm Trail",{{	IDR_CHISHOLMG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	IDR_CHISHOLMC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Football",{	{	IDR_FOOTBALLG,	0x6000, 0x4000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Hustle",	{	{	IDR_HUSTLEG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Hunt the Wumpus",{{IDR_WUMPUSG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},
 
	{	"Mind Challengers",{{IDR_MINDG,		0x6000, 0x2000,	TYPE_GROM	, 0},
						{	0,				0x6000,	0x2000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Munch Man",{	{	IDR_MUNCHMNG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	IDR_MUNCHMNC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},
	
	{	"Parsec",	{	{	IDR_PARSECG,	0x6000, 0x6000,	TYPE_GROM	, 0},
						{	IDR_PARSECC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},


	{	"Super Space Acer",{{IDR_SSALOAD,	0x6000, 0x0030,	TYPE_ROM	, 0},
						{	IDR_DEMQ,		0x2000,	0x0706,	TYPE_RAM	, 0},
						{	IDR_SSARAM,		0xA000,	0x5A28,	TYPE_RAM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		Disk_SSA,
		NULL
	},

	{	"TI Invaders",{	{	IDR_TIINVADG,	0x6000, 0x8000,	TYPE_GROM	, 0},
						{	IDR_TIINVADC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},
	
#if 0
	// not working yet
	{	"TI-Nopoly",{	{	IDR_ROMS4,		0x6000,	0x0030,	TYPE_ROM	, 0},
						{	IDR_ROMS1,		0x0003, 0x0df2,	TYPE_AMS	, 0},	// both words used for size on AMS cards
						{	IDR_ROMS2,		0x2000,	0x00BA,	TYPE_RAM	, 0},
						{	IDR_ROMS6,		0x0000,	0x0005,	TYPE_KEYS	, 0},
					},
		NULL,
		"AMS Card will automatically be enabled."
	},
#endif

	{	"Tombstone City",{{	IDR_TOMBCITG,	0x6000, 0x2000,	TYPE_GROM	, 0},
						{	IDR_TOMBCITC,	0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},

	{	"Tunnels of Doom",{{ IDR_TUNDOOMG,	0x6000,	0xA000,	TYPE_GROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		Disk_Tunnels,
		"Select DSK1, and PENNY for introductory quest, or QUEST for a full quest."
	},

	{	"Video Chess",{	{	IDR_CHESSG,		0x6000, 0x8000,	TYPE_GROM	, 0},
						{	IDR_CHESSC,		0x6000,	0x2000,	TYPE_ROM	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
						{	0,				0x0000,	0x0000,	TYPE_NONE	, 0},
					},
		NULL,
		NULL
	},
};

// Configuration access
void ReadConfig() {
	int idx,idx2,idx3;

	for (idx=0; idx<3; idx++) {
		fVoiceSimulatedDAC[idx]=false;
	}

	// type of tone: 0-9 (0=default square)
	nSampleIdx[0]=			GetPrivateProfileInt("audio",	"sampleidx0",	nSampleIdx[0],				INIFILE);
	nSampleIdx[1]=			GetPrivateProfileInt("audio",	"sampleidx1",	nSampleIdx[1],				INIFILE);
	nSampleIdx[2]=			GetPrivateProfileInt("audio",	"sampleidx2",	nSampleIdx[2],				INIFILE);
	fSampleShouldLoop[0]=(	GetPrivateProfileInt("audio",	"sampleloop0",	fSampleShouldLoop[0]?1:0,	INIFILE)!=0);
	fSampleShouldLoop[1]=(	GetPrivateProfileInt("audio",	"sampleloop1",	fSampleShouldLoop[1]?1:0,	INIFILE)!=0);
	fSampleShouldLoop[2]=(	GetPrivateProfileInt("audio",	"sampleloop2",	fSampleShouldLoop[2]?1:0,	INIFILE)!=0);
	// Volume multiplier
	volume_mult=			GetPrivateProfileInt("audio",	"volume_mult",	volume_mult,				INIFILE);

	// Disk paths
	for (idx=0; idx<10; idx++) {
		char buf[128];
		sprintf(buf, "dskpath%d", idx);
		// Path to the disk files - usually relative but not required
		GetPrivateProfileString("disk", buf, diskpath[idx],	diskpath[idx], 256, INIFILE);
	}
	// type of files in folder - soon to be deprecated: T-TIFILES, V-V9T9 files, I-DSK image, R-Raw files
	// (Only TIFILES and V9T9 files are written, raw support to be removed eventually)
	// (One char per disk, 10 chars max)
	GetPrivateProfileString("disk",	"types", disktype, disktype, 11, INIFILE);

	// Filename used to write recorded video
	GetPrivateProfileString("emulation", "AVIFilename", AVIFileName, AVIFileName, 256, INIFILE);
	// CPU Throttling? CPU_OVERDRIVE, CPU_NORMAL, CPU_MAXIMUM
	CPUThrottle=	GetPrivateProfileInt("emulation",	"cputhrottle",			CPUThrottle,	INIFILE);
	// VDP Throttling? VDP_CPUSYNC, VDP_REALTIME
	SystemThrottle=	GetPrivateProfileInt("emulation",	"systemthrottle",		SystemThrottle,	INIFILE);
	// Proper CPU throttle (cycles per frame) - ipf is deprecated
	max_cpf=		GetPrivateProfileInt("emulation",	"maxcpf",				max_cpf,		INIFILE);
	// Pause emulator when window inactive: 0-no, 1-yes
	PauseInactive=	GetPrivateProfileInt("emulation",	"pauseinactive",		PauseInactive,	INIFILE);
	// Disable speech cause it sucks at the moment
	SpeechEnabled=  GetPrivateProfileInt("emulation",   "speechenabled",         SpeechEnabled,  INIFILE);
	// Simulation of the floating point functions in the ROM: 0-no, 1-yes
	speedup_float=	GetPrivateProfileInt("emulation",	"speedup_float",		speedup_float,	INIFILE);
	// Simulate the console interrupt (faster): 0-no, 1-yes
	simulate_int=	GetPrivateProfileInt("emulation",	"speedup_interrupt",	simulate_int,	INIFILE);
	// Get system type: 0-99/4, 1-99/4A, 2-99/4Av2.2
	nSystem=		GetPrivateProfileInt("emulation",	"system",				nSystem,		INIFILE);
	// Read flag for slowing keyboard repeat: 0-no, 1-yes
	slowdown_keyboard=GetPrivateProfileInt("emulation",	"slowdown_keyboard",	slowdown_keyboard, INIFILE);
	// Check whether to use the ps/2 keyboard (normally yes for 99/4A)
	if (nSystem == 0) {
		keyboard=KEY_994;		// 99/4
	} else {
		ps2keyboardok=GetPrivateProfileInt("emulation", "ps2keyboard", 1, INIFILE);
		if (ps2keyboardok) {
			keyboard=KEY_994A_PS2;	// 99/4A with ps/2
		} else {
			keyboard=KEY_994A;		// 99/4A without ps/2
		}
	}
	// SAMS emulation
	sams_enabled = GetPrivateProfileInt("emulation", "sams_enabled", sams_enabled, INIFILE);
	// Read flag for SAMS memory size if selected
	sams_size = GetPrivateProfileInt("emulation", "sams_size", sams_size, INIFILE);

	// Joystick active: 0 - off, 1 on
	fJoy=		GetPrivateProfileInt("joysticks", "active",		fJoy,		INIFILE);
	// 0-keyboard, 1-PC joystick 1, 2-PC joystick 2
	joy1mode=	GetPrivateProfileInt("joysticks", "joy1mode",	joy1mode,	INIFILE);
	joy2mode=	GetPrivateProfileInt("joysticks", "joy2mode",	joy2mode,	INIFILE);

	// Cartridge group loaded (0-apps, 1-games, 2-user)
	nCartGroup=	GetPrivateProfileInt("roms",	"cartgroup",	nCartGroup,	INIFILE);
	// Cartridge index (depends on group)
	nCart=		GetPrivateProfileInt("roms",	"cartidx",		nCart,		INIFILE);
	// User cartridges
	nLoadedUserCarts=0;
	idx=0;
	idx2=0;
	for (idx=0; idx<100; idx++) {
		char buf[256], buf2[256];

		sprintf(buf, "UserCart%d", idx);
		GetPrivateProfileString(buf, "name", "", Users[idx2].szName, 64, INIFILE);
		if (strlen(Users[idx2].szName) > 0) {
			Users[idx2].pDisk=NULL;
			GetPrivateProfileString(buf, "message", "", buf2, 256, INIFILE);
			if (strlen(buf2) > 0) {
				Users[idx2].szMessage=_strdup(buf2);		// this memory will leak!
			} else {
				Users[idx2].szMessage=NULL;
			}
			for (idx3=0; idx3<MAXROMSPERCART; idx3++) {
				char buf3[1024];

				sprintf(buf2, "ROM%d", idx3);
				// line is formatted, except filename which finishes the line
				// T[x]|AAAA|LLLL|filename
				// [x] is the optional bank number from 0-F
				Users[idx2].Img[idx3].dwImg=NULL;
				Users[idx2].Img[idx3].nBank=0;
				GetPrivateProfileString(buf, buf2, "", buf3, 1024, INIFILE);
				if (strlen(buf3) > 0) {
					if (4 != sscanf(buf3, "%c|%x|%x|%s", 
						&Users[idx2].Img[idx3].nType,
						&Users[idx2].Img[idx3].nLoadAddr,
						&Users[idx2].Img[idx3].nLength,
						&Users[idx2].Img[idx3].szFileName)) {
							if (5 != sscanf(buf3, "%c%x|%x|%x|%s", 
								&Users[idx2].Img[idx3].nType,
								&Users[idx2].Img[idx3].nBank,
								&Users[idx2].Img[idx3].nLoadAddr,
								&Users[idx2].Img[idx3].nLength,
								&Users[idx2].Img[idx3].szFileName)) {
									sprintf(buf3, "INI File error reading %s in %s", buf2, buf);
									MessageBox(NULL, buf3, "Classic99 Error", MB_OK);
									goto skiprestofuser;
							}
					}
					// this doesn't read correctly?
					Users[idx2].Img[idx3].nType=buf3[0];
				}
			}

			idx2++;
		}
	}
	nLoadedUserCarts=idx2;

skiprestofuser:
	// video filter mode
	FilterMode=		GetPrivateProfileInt("video",	"FilterMode",		FilterMode,		INIFILE);
	// essentially frameskip
	drawspeed=		GetPrivateProfileInt("video",	"frameskip",		drawspeed,		INIFILE);
	// graphics mode used for full screen direct X (see SetupDirectDraw() in tivdp.cpp)
	FullScreenMode=	GetPrivateProfileInt("video",	"fullscreenmode",	FullScreenMode, INIFILE);
	// set interrupt rate - 50/60
	hzRate=			GetPrivateProfileInt("video",	"hzRate",			hzRate,			INIFILE);
	if ((hzRate != 50) && (hzRate != 60)) {
		hzRate=60;
	}
	// whether to force correct aspect ratio
	MaintainAspect=	GetPrivateProfileInt("video",	"MaintainAspect",	MaintainAspect, INIFILE);
	// 0-none, 1-DIB, 2-DX, 3-DX Full
	StretchMode=	GetPrivateProfileInt("video",	"StretchMode",		StretchMode,	INIFILE);

	// TV stuff
	TVScanLines=	GetPrivateProfileInt("tvfilter","scanlines",		TVScanLines,	INIFILE);
	double thue, tsat, tcont, tbright, tsharp, tmp;
	tmp=			GetPrivateProfileInt("tvfilter","hue",				100,			INIFILE);
	thue=(tmp-100)/100.0;
	tmp=			GetPrivateProfileInt("tvfilter","saturation",		100,			INIFILE);
	tsat=(tmp-100)/100.0;
	tmp=			GetPrivateProfileInt("tvfilter","contrast",			100,			INIFILE);
	tcont=(tmp-100)/100.0;
	tmp=			GetPrivateProfileInt("tvfilter","brightness",		100,			INIFILE);
	tbright=(tmp-100)/100.0;
	tmp=			GetPrivateProfileInt("tvfilter","sharpness",		100,			INIFILE);
	tsharp=(tmp-100)/100.0;
	SetTVValues(thue, tsat, tcont, tbright, tsharp);
}

// Wrapper function - not available in Win32?
void WritePrivateProfileInt(LPCTSTR lpApp, LPCTSTR lpKey, int nVal, LPCTSTR lpFile) {
	char buf[256];

	sprintf(buf, "%d", nVal);
	WritePrivateProfileString(lpApp, lpKey, buf, lpFile);
}

void SaveConfig() {
	int idx;

	WritePrivateProfileInt(		"audio",		"sampleidx0",			nSampleIdx[0],				INIFILE);
	WritePrivateProfileInt(		"audio",		"sampleidx1",			nSampleIdx[1],				INIFILE);
	WritePrivateProfileInt(		"audio",		"sampleidx2",			nSampleIdx[2],				INIFILE);
	WritePrivateProfileInt(		"audio",		"sampleloop0",			fSampleShouldLoop[0]?1:0,	INIFILE);
	WritePrivateProfileInt(		"audio",		"sampleloop1",			fSampleShouldLoop[1]?1:0,	INIFILE);
	WritePrivateProfileInt(		"audio",		"sampleloop2",			fSampleShouldLoop[2]?1:0,	INIFILE);
	WritePrivateProfileInt(		"audio",		"volume_mult",			volume_mult,				INIFILE);

	for (idx=0; idx<10; idx++) {
		char buf[128];
		sprintf(buf, "dskpath%d", idx);
		WritePrivateProfileString("disk",		buf,					diskpath[idx],				INIFILE);
	}
	WritePrivateProfileString(	"disk",			"types",				disktype,					INIFILE);

	WritePrivateProfileString(	"emulation",	"AVIFilename",			AVIFileName,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"cputhrottle",			CPUThrottle,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"systemthrottle",		SystemThrottle,				INIFILE);
	if (0 != max_cpf) {
		WritePrivateProfileInt(	"emulation",	"maxcpf",				max_cpf,					INIFILE);
	}
	WritePrivateProfileInt(		"emulation",	"pauseinactive",		PauseInactive,				INIFILE);
	WritePrivateProfileInt(     "emulation",    "speechenabled",        SpeechEnabled,              INIFILE);
	WritePrivateProfileInt(		"emulation",	"speedup_float",		speedup_float,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"speedup_interrupt",	simulate_int,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"system",				nSystem,					INIFILE);
	WritePrivateProfileInt(		"emulation",	"slowdown_keyboard",	slowdown_keyboard,			INIFILE);
	WritePrivateProfileInt(		"emulation",	"ps2keyboard",			ps2keyboardok,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"sams_enabled",			sams_enabled,				INIFILE);
	WritePrivateProfileInt(		"emulation",	"sams_size",			sams_size,					INIFILE);

	WritePrivateProfileInt(		"joysticks",	"active",				fJoy,						INIFILE);
	WritePrivateProfileInt(		"joysticks",	"joy1mode",				joy1mode,					INIFILE);
	WritePrivateProfileInt(		"joysticks",	"joy2mode",				joy2mode,					INIFILE);

	WritePrivateProfileInt(		"roms",			"cartgroup",			nCartGroup,					INIFILE);
	WritePrivateProfileInt(		"roms",			"cartidx",				nCart,						INIFILE);
	
	WritePrivateProfileInt(		"video",		"FilterMode",			FilterMode,					INIFILE);
	WritePrivateProfileInt(		"video",		"frameskip",			drawspeed,					INIFILE);
	WritePrivateProfileInt(		"video",		"fullscreenmode",		FullScreenMode,				INIFILE);
	WritePrivateProfileInt(		"video",		"hzRate",				hzRate,						INIFILE);
	WritePrivateProfileInt(		"video",		"MaintainAspect",		MaintainAspect,				INIFILE);
	WritePrivateProfileInt(		"video",		"StretchMode",			StretchMode,				INIFILE);

	// TV stuff
	double thue, tsat, tcont, tbright, tsharp;
	int tmp;
	GetTVValues(&thue, &tsat, &tcont, &tbright, &tsharp);

	tmp=(int)((thue+1.0)*100.0);
	WritePrivateProfileInt(		"tvfilter",		"hue",					tmp,						INIFILE);
	tmp=(int)((tsat+1.0)*100.0);
	WritePrivateProfileInt(		"tvfilter",		"saturation",			tmp,						INIFILE);
	tmp=(int)((tcont+1.0)*100.0);
	WritePrivateProfileInt(		"tvfilter",		"contrast",				tmp,						INIFILE);
	tmp=(int)((tbright+1.0)*100.0);
	WritePrivateProfileInt(		"tvfilter",		"brightness",			tmp,						INIFILE);
	tmp=(int)((tsharp+1.0)*100.0);
	WritePrivateProfileInt(		"tvfilter",		"sharpness",			tmp,						INIFILE);
	WritePrivateProfileInt(		"tvfilter",		"scanlines",			TVScanLines,				INIFILE);
}

// convert config into meaningful values for AMS system
void SetupSams(int sams_mode, int sams_size) {
	EmulationMode emuMode = None;
	AmsMemorySize amsSize = Mem128k;
	
	// We don't really NEED this translation layer, but we can remove it later.
	if (sams_mode) {
		// currently only SuperAMS, so if anything set use that
		emuMode = Sams;

		switch (sams_size) {
		case 1:
			amsSize = Mem256k;
			break;
		case 2:
			amsSize = Mem512k;
			break;
		case 3:
			amsSize = Mem1024k;
			break;
		default:
			break;
		}
	}

	SetAmsMemorySize(amsSize);
	InitializeMemorySystem(emuMode);	
}

///////////////////////////////////
// Main
// Startup and shutdown system
///////////////////////////////////
int WINAPI WinMain( HINSTANCE hInst, HINSTANCE hInPrevInstance, LPSTR /*lpCmdLine*/, int /*nCmdShow*/)
{
	int idx;
	int err;
	char temp[255];
	WNDCLASS aclass;
	TEXTMETRIC myMetric;
	RECT myrect, myrect2;

	hInstance = hInst;
	hPrevInstance=hInPrevInstance;

	// Null the pointers
	myClass=0;
	myWnd=NULL;		// Classic99 Window
	dbgWnd=NULL;	// Debug Window
	lpds=NULL;
	voice[0]=NULL;
	voice[1]=NULL;
	voice[2]=NULL;
	PasteString=NULL;
	PasteIndex=NULL;
	PasteCount=0;
	ZeroMemory(Users, sizeof(Users));
	nLoadedUserCarts=0;
	memset(BreakPoint, 0, sizeof(BreakPoint));
	memset(BreakPointType, 0, sizeof(BreakPointType));
	nBreakPoints=0;

	// Get the default np font dimensions with a dummy dc
	myDC=CreateCompatibleDC(NULL);
	SelectObject(myDC, GetStockObject(ANSI_FIXED_FONT));
	if (GetTextMetrics(myDC, &myMetric)) {
		fontX=myMetric.tmMaxCharWidth;
		fontY=myMetric.tmHeight;
	} else {
		fontX=20;
		fontY=20;
	}
	DeleteDC(myDC);

	for (idx=0; idx<8; idx++) {
		noise[idx]=NULL;
	}

	framedata=(unsigned short*)malloc((256+16)*(192+16)*2);	// This is where we draw everything - 8 pixel border
	framedata2=(unsigned short*)malloc((640)*(384+32)*2);// used for the 2xSaI and TV filters - 16 pixel border on SAI and 8 horizontal on TV (x2)

	MyMutex=CreateMutex(NULL, false, NULL);

	// create and register a class and open a window
	if (NULL == hPrevInstance)
	{
		aclass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
		aclass.lpfnWndProc = myproc;
		aclass.cbClsExtra = 0;
		aclass.cbWndExtra = 0;
		aclass.hInstance = hInstance;
		aclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
		aclass.hCursor = LoadCursor(NULL, IDC_ARROW);
		aclass.hbrBackground = NULL;
		aclass.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
		aclass.lpszClassName = "TIWndClass";
		myClass = RegisterClass(&aclass);
		if (0 == myClass)
		{	
			err=GetLastError();
			sprintf(temp, "Can't create class: 0x%x", err);
			fail(temp);
		}

		// Info windows
		aclass.lpszClassName="Classic99Info";
		aclass.hbrBackground=NULL;
		aclass.lpszMenuName=NULL;
		myClass=RegisterClass(&aclass);
		if (0 == myClass) {
			debug_write("Couldn't register debug window class: 0x%x", GetLastError());
		}
	}

	myWnd = CreateWindow("TIWndClass", "Classic99", WS_OVERLAPPEDWINDOW | WS_SIZEBOX | WS_VISIBLE, CW_USEDEFAULT, 0, 536, 446, NULL, NULL, hInstance, NULL);
	if (NULL == myWnd)
	{	
		err=GetLastError();
		sprintf(temp, "Can't open window: %x", err);
		fail(temp);
	}

	// start the debug updater thread
	if (-1 == _beginthread(DebugUpdateThread, 0, NULL)) {
		debug_write("Failed to start debug update thread.");
	} else {
		debug_write("Debug update thread started.");
	}

	// Fill in the menus
	HMENU hMenu;
	// Systems
	hMenu=GetMenu(myWnd);
	if (hMenu) {
		hMenu=GetSubMenu(hMenu, 2);
		if (hMenu) {
			for (idx=0; idx<sizeof(Systems)/sizeof(struct CARTS); idx++) {
				AppendMenu(hMenu, MF_STRING, ID_SYSTEM_0+idx, Systems[idx].szName);
			}
			DeleteMenu(hMenu, 0, MF_BYPOSITION);	// remove temp separator
		}
	}
	// Apps
	hMenu=GetMenu(myWnd);
	if (hMenu) {
		hMenu=GetSubMenu(hMenu, 3);
		if (hMenu) {
			hMenu=GetSubMenu(hMenu, 0);
			if (hMenu) {
				for (idx=0; idx<sizeof(Apps)/sizeof(struct CARTS); idx++) {
					AppendMenu(hMenu, MF_STRING, ID_APP_0+idx, Apps[idx].szName);
				}
				DeleteMenu(hMenu, 0, MF_BYPOSITION);	// remove temp separator
			}
		}
	}
	// Games
	hMenu=GetMenu(myWnd);
	if (hMenu) {
		hMenu=GetSubMenu(hMenu, 3);
		if (hMenu) {
			hMenu=GetSubMenu(hMenu, 1);
			if (hMenu) {
				for (idx=0; idx<sizeof(Games)/sizeof(struct CARTS); idx++) {
					AppendMenu(hMenu, MF_STRING, ID_GAME_0+idx, Games[idx].szName);
				}
				DeleteMenu(hMenu, 0, MF_BYPOSITION);	// remove temp separator
			}
		}
	}
	ShowWindow(myWnd, SW_SHOWNORMAL);

	// Get window trimming size
	GetWindowRect(myWnd, &myrect);
	GetClientRect(myWnd, &myrect2);
	wndXtrim=((myrect.right - myrect.left)-(myrect2.right - myrect2.left))-1;
	wndYtrim=((myrect.bottom - myrect.top)-(myrect2.bottom - myrect2.top))-fontY;

	for (idx=0; idx<=PCODEGROMBASE; idx++) {
		GROMBase[idx].GRMADD=0;
		GROMBase[idx].grmaccess=2;
		GROMBase[idx].grmdata=0;
	}
		
	vdpaccess=0;		// No VDP address writes yet 
	vdpprefetch=0;		// Not really accurate, but eh
	tmpVDPADD=0;		// part of the vdpaccess
	interrupt_needed=0;	// No interrupt missed yet
	quitflag=0;			// no quit yet
	nCurrentDSR=-1;		// no DSR selected
	memset(nDSRBank, 0, sizeof(nDSRBank));
	timer9901=0;		// timer is idle
	starttimer9901=0;
	timer9901IntReq=0;
	timer9901IntOk=0;

	// clear debugging strings
	memset(lines, 0, sizeof(lines));
	memset(disasm, 0, sizeof(disasm));
	bDebugDirty=true;

	//DacLog=fopen("C:\\daclog.raw", "wb");

	// Print some initial debug
	debug_write("---");
	debug_write("Classic99 version %s (C)2002-2009 M.Brent", VERSION);
	debug_write("ROM files included under license from Texas Instruments");

	// Set default values for config (alphabetized here)
	strcpy(AVIFileName, "C:\\Classic99.AVI");	// default movie filename
	nCartGroup=0;				// Cartridge group (0-apps, 1-games, 2-user)
	nCart=-1;					// loaded cartridge (-1 is none)
	CPUThrottle=CPU_NORMAL;		// throttle the CPU
	SystemThrottle=VDP_CPUSYNC;	// throttle the VDP
	drawspeed=0;				// no frameskip
	FilterMode=2;				// super 2xSAI
	FullScreenMode=6;			// full screen at 640x480x16
	fJoy=1;						// enable joysticks
	joy1mode=0;					// keyboard
	joy2mode=1;					// joystick 1
	hzRate=60;					// 60 hz
	MaintainAspect=1;			// Keep aspect ratio
	max_cpf=DEFAULT_CPF;		// max cycles per frame
	PauseInactive=0;			// don't pause when window inactive
	SpeechEnabled=1;			// speech is decent now
	Recording=0;				// not recording AVI
	nSampleIdx[0]=0;			// square wave
	nSampleIdx[1]=0;			// square wave
	nSampleIdx[2]=0;			// square wave
	fSampleShouldLoop[0]=true;	// default for square wave
	fSampleShouldLoop[1]=true;	// default for square wave
	fSampleShouldLoop[2]=true;	// default for square wave
#ifdef ALLOW_OLD_HACKS
	simulate_int=0;				// don't use interrupt simulator
	speedup_float=0;			// don't use float simulator
#endif
	slowdown_keyboard=1;		// slow down keyboard repeat when read via BASIC
	StretchMode=2;				// dx
	TVScanLines=1;				// on by default
	sams_enabled=1;				// off by default
	sams_size=3;				// 1MB by default when on (no reason not to use a large card)
	nSystem=1;					// TI-99/4A
	volume_mult=-200;			// Default volume attenuation
	CPUSpeechHalt=false;		// not halted for speech reasons
	CPUSpeechHaltByte=0;		// doesn't matter

	// Default to only DSK1 setup as a folder
	ZeroMemory(diskpath, sizeof(diskpath));
	strcpy(disktype, "..........");		// must be 10 dots
	strcpy(&diskpath[1][0], "DSK1\\");
	disktype[1] = 'T';

	// Read configuration - uses above settings as default!
	ReadConfig();

	// Update user menu - this will be a function later
	hMenu=GetMenu(myWnd);
	if (hMenu) {
		hMenu=GetSubMenu(hMenu, 3);
		if (hMenu) {
			hMenu=GetSubMenu(hMenu, 2);
			if (hMenu) {
				// User ROMs are a bit different, since we have to read them
				// from the configuration
				for (idx=0; idx<nLoadedUserCarts; idx++) {
					AppendMenu(hMenu, MF_STRING, ID_USER_0+idx, Users[idx].szName);
				}
			}
		}
	}
	
	// set temp stuff
	oldmax=max_cpf;

	// set up SAMS emulation
	SetupSams(sams_enabled, sams_size);

	// Load a dummy CPU ROM for the emu to spin on till we load something real
	//memcpy(CPU, DummyROM, 6);
	WriteMemoryBlock(0x0000, DummyROM, 6);

	// Set menu-based settings (lParam 1 means it's coming from here, not the user)
	// Only some messages care about that param, though
	PostMessage(myWnd, WM_COMMAND, ID_SYSTEM_0+nSystem, 1);
	PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
	PostMessage(myWnd, WM_COMMAND, ID_VIDEO_MAINTAINASPECT, 1);
	PostMessage(myWnd, WM_COMMAND, ID_VIDEO_FILTERMODE_NONE+FilterMode, 1);

	if ((StretchMode>=0)&&(StretchMode<=2)) {
		PostMessage(myWnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_NONE+StretchMode, 1);
	} else {
		if (StretchMode==3) {
			PostMessage(myWnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_DXFULL_320X240X8+FullScreenMode-1, 1);
		}
	}
	PostMessage(myWnd, WM_COMMAND, ID_VIDEO_50HZ, 1);
	PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_PAUSEINACTIVE, 1);
	PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_SPEECHENABLED, 1);
	
	if (nCart != -1) {
		switch (nCartGroup) {
		case 0:
			PostMessage(myWnd, WM_COMMAND, ID_APP_0+nCart, 1);
			break;

		case 1:
			PostMessage(myWnd, WM_COMMAND, ID_GAME_0+nCart, 1);
			break;

		case 2:
			PostMessage(myWnd, WM_COMMAND, ID_USER_0+nCart, 1);
			break;
		}
	}

	// build the CPU table
	debug_write("Building CPU");
	buildcpu();
	hWakeupEvent=CreateEvent(NULL, FALSE, FALSE, NULL);
	InterlockedExchange((LONG*)&cycles_left, max_cpf);

	// Initialize emulated keyboard
	init_kb();

	// start sound
	debug_write("Starting Sound");
	startsound();
	
	// Init disk
	debug_write("Starting Disk");
	do_files(3);		// TI DSR sets up VDP for 3 files

	// prepare the emulation...
	cpucount=0;
	cpuframes=0;
	timercount=0;

	// start the video processor
	debug_write("Starting Video");
	startvdp();

	// set up 60hz timer
	myThread=_beginthread(TimerThread, 0, NULL);
	if (myThread != -1) {
		debug_write("Timer thread began...");
	} else {
		debug_write("Timer thread failed.");
	}
	// set up speech buffer restart
	myThread=_beginthread(SpeechBufThread, 0, NULL);
	if (myThread != -1) {
		debug_write("Speech Buffer thread began...");
	} else {
		debug_write("Speech Buffer thread failed.");
	}

	Sleep(100);			// time for threads to start

	// begin emulation - returns when it's time to exit
	debug_write("Starting Emulation");
	emulti();

	// save out our config
	SaveConfig();

	// Fail is the full exit
	debug_write("Shutting down");
	fail("Normal Termination");
	CloseHandle(hWakeupEvent);

	ShutdownMemorySystem();

	// good bye
	return 0;
}


//////////////////////////////////////////////////////
// start up the sound system
//////////////////////////////////////////////////////
bool LoadResourceWAV(LPDIRECTSOUNDBUFFER *ppdSnd, int nResource) {
	unsigned char *pData;
	HRSRC hRsrc;
	HGLOBAL hGlob;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX pcmwf;
	unsigned int idx, idx2;
	UCHAR *ptr1, *ptr2;
	unsigned long len1, len2, len;
	char buf[80];

	ZeroMemory(&pcmwf, sizeof(pcmwf));
	hRsrc=NULL;

	hRsrc=FindResource(NULL, MAKEINTRESOURCE(nResource), "WAVE"); 
	if (hRsrc) {
		hGlob=LoadResource(NULL, hRsrc);
		if (NULL != hGlob) {
			pData=(unsigned char*)LockResource(hGlob);
			if (NULL != pData) {
				// in a demonstration of bad programming, I'm just going to read the darn thing ;)
				// if it's not a WAV, you may be in trouble. Even if it IS a WAV, you may be. ;)
				// This is hard coded to the WAVs I inserted into the resources.

				// assume data chunk starts at 0x24
				idx2=0x2c;								// first byte of data

				// it's simple. I expect it to be a 22khz, 8 bit mono WAV with 1 data part
				// good luck with anything else
				len=pData[0x28]+(256*pData[0x29])+(65536*pData[0x2a]); // and we'll ignore the highest byte

				// we could read this data now instead of assuming?
				pcmwf.wFormatTag = WAVE_FORMAT_PCM;		// wave file
				pcmwf.nChannels=1;						// 1 channel (mono)
				pcmwf.nSamplesPerSec=22050;				// 22khz
				pcmwf.nBlockAlign=1;					// 1 byte per sample * 1 channel
				pcmwf.nAvgBytesPerSec=pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
				pcmwf.wBitsPerSample=8;					// 8 bit samples
				pcmwf.cbSize=0;							// always zero;

				ZeroMemory(&dsbd, sizeof(dsbd));
				dsbd.dwSize=sizeof(dsbd);
				dsbd.dwFlags=DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC | DSBCAPS_GLOBALFOCUS;
				dsbd.dwBufferBytes=len;
				dsbd.lpwfxFormat=&pcmwf;

				if (FAILED(lpds->CreateSoundBuffer(&dsbd, ppdSnd, NULL)))
				{
					lpds->Release();
					return false;
				}
		
				if (SUCCEEDED((*ppdSnd)->Lock(0, len, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER)))
				{
					// blank the buffer
					memset(ptr1, 0, len1);

					// since we haven't started the sound, hopefully the second pointer is nil
					if (len2 != 0) {
						MessageBox(NULL, "Failed to lock sound buffer", "Classic99 Error", MB_OK);
					}

					memcpy(ptr1, &pData[idx2], min(len, len1));
				}
				(*ppdSnd)->Unlock(ptr1, len1, ptr2, len2);
			}
		}
		return true;
	} else {
		return false;
	}
}

void GenerateToneBuffer(int idx) {
	unsigned int idx2;
	UCHAR c;
	UCHAR *ptr1, *ptr2;
	unsigned long len1, len2, len;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX pcmwf;

	if (NULL != voice[idx]) {
		voice[idx]->Stop();
		voice[idx]->Release();
	}

	// In case we need it below for square or sine waves, we'll fill in the structs
	ZeroMemory(&pcmwf, sizeof(pcmwf));
	pcmwf.wFormatTag = WAVE_FORMAT_PCM;		// wave file
	pcmwf.nChannels=1;						// 1 channel (mono)
	pcmwf.nSamplesPerSec= 22050;			// 22khz
	pcmwf.nBlockAlign=1;					// 1 byte per sample * 1 channel
	pcmwf.nAvgBytesPerSec=pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
	pcmwf.wBitsPerSample=8;					// 8 bit samples
	pcmwf.cbSize=0;							// always zero;

	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize=sizeof(dsbd);
	dsbd.dwFlags=DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC | DSBCAPS_GLOBALFOCUS;
	dsbd.dwBufferBytes=AUDIO_BUFFER_SIZE;	// the sample is AUDIO_BUFFER_SIZE bytes long
	dsbd.lpwfxFormat=&pcmwf;

	switch (nSampleIdx[idx]) {
	default:
	case 0:		// square wave
		if (FAILED(lpds->CreateSoundBuffer(&dsbd, &voice[idx], NULL)))
		{
			lpds->Release();
			return;
		}
		
		// I have calculated that the square wav changed state every 25 samples
		if (SUCCEEDED(voice[idx]->Lock(0, AUDIO_BUFFER_SIZE, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER)))
		{
			// since we haven't started the sound, hopefully the second pointer is null
			if (len2 != 0) {
				MessageBox(NULL, "Failed to lock tone buffer", "Classic99 Error", MB_OK);
			}

			// square wave with a period of 50 samples (441hz - but I assume 440 throughout)
			// unsigned
			c=200;
			for (idx2=0; idx2<len1; idx2++)
			{
				if ((idx2%25) == 0)
				{
					if (c==0) {
						c=200;
					} else {
						c=0;
					}
				}

				*(ptr1 + idx2)=c;
			}
			fSampleShouldLoop[idx]=true;

			voice[idx]->Unlock(ptr1, len1, ptr2, len2);
		}
		break;
	
	case 1:		// sine wave
		if (FAILED(lpds->CreateSoundBuffer(&dsbd, &voice[idx], NULL)))
		{
			lpds->Release();
			return;
		}
		
		// I have calculated that the square wav changed state every 25 samples
		if (SUCCEEDED(voice[idx]->Lock(0, AUDIO_BUFFER_SIZE, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER)))
		{
			// since we haven't started the sound, hopefully the second pointer is null
			if (len2 != 0) {
				MessageBox(NULL, "Failed to lock tone buffer", "Classic99 Error", MB_OK);
			}

			// Generate a sine wave with a period of 50 samples (441hz - but I assume 440 (A above Middle C) throughout)
			// since the musical scale is not exact, this is close enough. Middle C ranges from 256-280hz
			// depending on who you ask!
			for (idx2=0; idx2<len1; idx2++) {
				*(ptr1 + idx2) = (unsigned char)((int)((sin((6.283/50.0)*(idx2%50))*100)+100)&0xff);
			}
			fSampleShouldLoop[idx]=true;

			voice[idx]->Unlock(ptr1, len1, ptr2, len2);
		}
		break;

	case 2:		// flute
		LoadResourceWAV(&voice[idx], IDR_FLUTE);
		fSampleShouldLoop[idx]=false;
		break;

	case 3:		// oboe
		LoadResourceWAV(&voice[idx], IDR_OBOE);
		fSampleShouldLoop[idx]=false;
		break;

	case 4:		// organ
		LoadResourceWAV(&voice[idx], IDR_ORGAN);
		fSampleShouldLoop[idx]=false;
		break;

	case 5:		// piano
		LoadResourceWAV(&voice[idx], IDR_PIANO);
		fSampleShouldLoop[idx]=false;
		break;

	case 6:		// sax
		LoadResourceWAV(&voice[idx], IDR_SAX);
		fSampleShouldLoop[idx]=false;
		break;

	case 7:		// string
		LoadResourceWAV(&voice[idx], IDR_STRING);
		fSampleShouldLoop[idx]=false;
		break;

	case 8:		// trombone
		LoadResourceWAV(&voice[idx], IDR_TROMBONE);
		fSampleShouldLoop[idx]=false;
		break;

	case 9:		// violin
		LoadResourceWAV(&voice[idx], IDR_VIOLIN);
		fSampleShouldLoop[idx]=false;
		break;
	}
}

void startsound()
{ /* start up the sound files */

	DSBUFFERDESC dsbd;
	WAVEFORMATEX pcmwf;
	unsigned int idx, idx2;
	UCHAR *ptr1, *ptr2;
	unsigned long len1, len2, len;
	char buf[80];
	
	latch_byte=0;

	if (FAILED(DirectSoundCreate(NULL, &lpds, NULL)))
	{
		lpds=NULL;		// no sound
		return;
	}
	
	if (FAILED(lpds->SetCooperativeLevel(myWnd, DSSCL_NORMAL)))	// normal created a 22khz, 8 bit stereo DirectSound system
	{
		lpds->Release();
		return;
	}

	for (idx=0; idx<3; idx++)
	{
		GenerateToneBuffer(idx);

		if (FAILED(voice[idx]->SetVolume(DSBVOLUME_MIN))) {
			debug_write("Set volume failed");
		}

		if (FAILED(voice[idx]->Play(0, 0, fSampleShouldLoop[idx]?DSBPLAY_LOOPING:0))) {
			debug_write("Voice DID NOT START");
		}
	}

	// each noise channel now has it's own voice, to eliminate the need to change buffers
	// and mess with repeat lengths and the like

	for (idx=0; idx<8; idx++)
	{
		bool fRet;

		switch (idx) {
		case 0: fRet=LoadResourceWAV(&noise[0], IDR_NOISE1); break;
		case 1: fRet=LoadResourceWAV(&noise[1], IDR_NOISE2); break;
		case 2: fRet=LoadResourceWAV(&noise[2], IDR_NOISE3); break;
		case 3: fRet=LoadResourceWAV(&noise[3], IDR_NOISE4); break;
		case 4: fRet=LoadResourceWAV(&noise[4], IDR_NOISE5); break;
		case 5: fRet=LoadResourceWAV(&noise[5], IDR_NOISE6); break;
		case 6: fRet=LoadResourceWAV(&noise[6], IDR_NOISE7); break;
		case 7: fRet=LoadResourceWAV(&noise[7], IDR_NOISE8); break;
		}
		
		if (fRet) {
			if (FAILED(noise[idx]->SetVolume(DSBVOLUME_MIN))) {
				debug_write("Set volume failed");
			}

			if (FAILED(noise[idx]->Play(0, 0, DSBPLAY_LOOPING))) {
				debug_write("Noise DID NOT START");
			}
		}
	}

	// load the Speech DLL
	hSpeechDll=LoadLibrary("SpeechDll.dll");
	if (NULL == hSpeechDll) {
		debug_write("Failed to load speech library.");
	} else {
		SpeechInit=(void (*)(Byte*,int,int,int))GetProcAddress(hSpeechDll, "SpeechInit");
		SpeechStop=(void (*)(void))GetProcAddress(hSpeechDll, "SpeechStop");
		SpeechRead=(Byte (*)(void))GetProcAddress(hSpeechDll, "SpeechRead");
		SpeechWrite=(bool (*)(Byte, bool))GetProcAddress(hSpeechDll, "SpeechWrite");
		SpeechProcess=(void (*)(Byte*,int))GetProcAddress(hSpeechDll, "SpeechProcess");
	}

	// Zero the temporary buffers
	memset(SpeechTmp, 0, sizeof(SpeechTmp));
	nSpeechTmpPos=0;
	memset(DACTmp, 0, sizeof(DACTmp));
	nDACPos=0;

	// Now load up the speech system
	/* start audio stream - SPEECHBUFFER buffer, 16 bit, 8khz, max vol, center */
	ZeroMemory(&pcmwf, sizeof(pcmwf));
	pcmwf.wFormatTag = WAVE_FORMAT_PCM;		// wave file
	pcmwf.nChannels=1;						// 1 channel (mono)
	pcmwf.nSamplesPerSec=SPEECHRATE;		// Should be 8khz
	pcmwf.nBlockAlign=2;					// 2 bytes per sample * 1 channel
	pcmwf.nAvgBytesPerSec=pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
	pcmwf.wBitsPerSample=16;				// 16 bit samples
	pcmwf.cbSize=0;							// always zero;

	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize=sizeof(dsbd);
	dsbd.dwFlags=DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_GLOBALFOCUS;
	dsbd.dwBufferBytes=SPEECHBUFFER;
	dsbd.lpwfxFormat=&pcmwf;

	hSpeechBufferClearEvent=CreateEvent(NULL, FALSE, FALSE, NULL);

	if (FAILED(lpds->CreateSoundBuffer(&dsbd, &speechbuf, NULL))) {
		debug_write("Failed to create speech sound buffer!");
	} else {
		if (SUCCEEDED(speechbuf->Lock(0, SPEECHBUFFER*pcmwf.nBlockAlign, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER))) {
			// since we haven't started the sound, hopefully the second pointer is nil
			if (len2 != 0) {
				debug_write("Failed to lock speech buffer");
			}

			// signed 16-bit - zero the buffer
			memset(ptr1, 0, len1);

			speechbuf->Unlock(ptr1, len1, ptr2, len2);
		}

		LPDIRECTSOUNDNOTIFY8 lpDsNotify;
		if (SUCCEEDED(speechbuf->QueryInterface(IID_IDirectSoundNotify8,(LPVOID*)&lpDsNotify))) {
			DSBPOSITIONNOTIFY notify;
			notify.dwOffset=DSBPN_OFFSETSTOP;
			notify.hEventNotify=hSpeechBufferClearEvent;
			if (FAILED(lpDsNotify->SetNotificationPositions(1, &notify))) {
				debug_write("Failed to set notification positions.");
			}
			lpDsNotify->Release();
		} else {
			debug_write("Failed to get DS8 interface for speech.");
		}

		if (FAILED(speechbuf->Play(0, 0, 0))) {
			debug_write("Speech DID NOT START");
		}

		if (SpeechInit) SpeechInit(SPEECH, 0x8000, SPEECHBUFFER, SPEECHRATE);
	}
}

//////////////////////////////////////////////////////////
// Start up the video system 
//////////////////////////////////////////////////////////
void startvdp()
{ 
	// call VDP Startup
	hVideoThread=NULL;
	hVideoThread=_beginthread((void (__cdecl *)(void*))VDPmain, 0, NULL);
	if (hVideoThread != -1)
		debug_write("Video Thread began...");
	else
		debug_write("Video Thread failed.");

	Sleep(100);

	// first retrace
	retrace_count=0;
}

//////////////////////////////////////////////////////////
// Non-fatal recoverable (?) error
//////////////////////////////////////////////////////////
void warn(char *x)
{ 
	// Warn will for now just dump a message into the log
	// eventually it should pop up a window and ask about
	// continuing

	debug_write("%s at address >%04x", x, PC);
}

//////////////////////////////////////////////////////////
// Fatal error - clean up and exit
// Note that normal exit is a fatal error ;)
//////////////////////////////////////////////////////////
void fail(char *x)
{ 
	// fatal error
	char buffer[1024];
	char buf2[256];
	int idx;

	// just in case it's not set yet
	quitflag=1;

	// add to the log - not useful now, but maybe in the future when it writes to disk
	debug_write(x);

	timeEndPeriod(1);

	sprintf(buffer,"\n%s\n",x);
	sprintf(buf2,"PC-%.4X  WP-%.4X  ST-%.4X  OP-%.4X\nGROM-%.4X VDP-%.4X\n",PC,WP,ST,in,GROMBase[0].GRMADD,VDPADD);
	strcat(buffer,buf2);
	sprintf(buf2,"Run Duration  : %d seconds\n",timercount/60);
	strcat(buffer,buf2);
	sprintf(buf2,"Operation time: %d instructions processed.\n",cpucount);
	strcat(buffer,buf2);
	sprintf(buf2,"Display frames: %d video frames displayed.\n",cpuframes);
	strcat(buffer,buf2);

	if (timercount<60) timercount=60;	// avoid divide by zero
	
	sprintf(buf2,"Average speed : %d instructions per second.\n",cpucount/(timercount/60));
	strcat(buffer,buf2);
	sprintf(buf2,"Frameskip     : %d\n",drawspeed);
	strcat(buffer,buf2);

	// the messagebox fails during a normal exit in WIN32.. why is that?
	OutputDebugString(buffer);
	MessageBox(myWnd, buffer, "Classic99 Exit", MB_OK);

	Sleep(600);			// give the threads a little time to shut down

	if (Recording) {
		CloseAVI();
	}

	if (SpeechStop) SpeechStop();

    if (speechbuf) {
		speechbuf->Stop();
		speechbuf->Release();
		speechbuf=NULL;
	}

	for (idx=0; idx<3; idx++) {
		if (voice[idx])
		{
			voice[idx]->Stop();
			voice[idx]->Release();
		}
	}
	for (idx=0; idx<8; idx++) {
		if (noise[idx])
		{
			noise[idx]->Stop();
			noise[idx]->Release();
		}
	}

	if (lpds) {
		lpds->Release();
	}

	if (myWnd) {
		DestroyWindow(myWnd);
	}
	
	if (myClass) {
		UnregisterClass("TIWndClass", hInstance);
	}

	if (framedata) free(framedata);
	if (framedata2) free(framedata2);

	if (hSpeechDll) {
		FreeLibrary(hSpeechDll);
		hSpeechDll=NULL;
	}

	exit(0);
}

/////////////////////////////////////////////////////////
// Return a Word from CPU memory
/////////////////////////////////////////////////////////
Word romword(Word x, bool rmw)
{ 
	x&=0xfffe;		// drop LSB
	return((rcpubyte(x,rmw)<<8)+rcpubyte(x+1,rmw));
}
/////////////////////////////////////////////////////////
// Return a Word from CPU memory using byte access
/////////////////////////////////////////////////////////
Byte romwordbyte(Word x)
{ 
	return rcpubyte(x);
}

/////////////////////////////////////////////////////////
// Write a Word to CPU memory
/////////////////////////////////////////////////////////
void wrword(Word x, Word y)
{ 
	Word nTmp;

	x&=0xfffe;		// drop LSB

	// now write the new data
	wcpubyte(x,(Byte)(y>>8));
	wcpubyte(x+1,(Byte)(y&0xff));

	// check breakpoints against what was written to where
	for (int idx=0; idx<nBreakPoints; idx++) {
		switch (BreakPointType[idx]&0xff) {
			case BREAK_EQUALS_WORD:
				if ((x == BreakPoint[idx]) && (y == ((BreakPointType[idx]&0xffff00)>>8))) {
					TriggerBreakPoint();
				}
				break;
			case BREAK_EQUALS_REGISTER:
				nTmp=WP+(BreakPoint[idx]*2);
				if ((x == nTmp) && (y == ((BreakPointType[idx]&0xffff00)>>8))) {
					TriggerBreakPoint();
				}
				break;
		}
	}
}

/////////////////////////////////////////////////////////
// Write a Byte to CPU memory using Word access
/////////////////////////////////////////////////////////
void wrwordbyte(Word x, Byte y)
{ 
	// now write the new data
	wcpubyte(x,y);
}

/////////////////////////////////////////////////////////
// Main loop for Emulation
/////////////////////////////////////////////////////////
void emulti()
{
	MSG msg;

	WP=romword(0);		// >0000 is the reset vector - read WP
	PC=romword(2);		// and then read the PC
	X_flag=0;			// not currently executing an X instruction
	quitflag=0;			// Don't quit
	nCycleCount=26;		// not that it's a big deal, but that's how long reset takes ;)

	while (!quitflag)
	{ 
		char buf[128];
		static FILE *fp=NULL;
		int cnt, idx, wid;

//	Full CPU trace to disk (warning: eats disk fast!)
//		cnt=Dasm9900(buf, PC);
//		if (NULL == fp) {
//			fp=fopen("C:\\9900.log", "w");
//		}
//		fprintf(fp, "%04x:", PC);
//		wid=18;
//		for (idx=0; idx<cnt; idx++) {
//			fprintf(fp, " %02x", GetSafeByte(PC+idx));
//			wid-=3;
//		}
//		fprintf(fp, "%*s", wid+1, " ");
//		fprintf(fp, ": %s\n", buf);
//

		// Dumps the XB Utilities to a disk file
//		if (0) {
//			fp=fopen("C:\\XB_UTILS", "wb");
//			fwrite(&CPU[0x2000], 1, 8192, fp);
//			fclose(fp);
//		}

		if ((PauseInactive)&&(myWnd != GetForegroundWindow())&&(dbgWnd != GetForegroundWindow())) {
			// we're supposed to pause when inactive, and we are not active
			// So, don't execute an instruction, and sleep a bit to relieve CPU
			Sleep(100);
		} else {
			// execute one opcode
			do1();
		}

#if 1
		extern Byte staticCPU[];
		//if (((staticCPU[0x83dd] == 0xb8)||(staticCPU[0x83dd] == 0xbe)) && (staticCPU[0x83dc] ==0x0a)) {
		if (bExtended) {
			SetWindowText(myWnd, "Classic99 - FORCED BREAK. F1 to Continue, F2 to Step");
			max_cpf=0;
			InterlockedExchange(&cycles_left, 0);
			draw_debug();
		}
		//staticCPU[0x836e]=0x4;
#endif

		// check for messages
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT) {
				quitflag=1;
#if 0
				{
					// dump scratchpad
					FILE *fp=fopen("C:\\ram.bin", "wb");
					fwrite(&staticCPU[0x8300], 1, 256, fp);
					fclose(fp);
				}
#endif
			}
			
			TranslateMessage(&msg);
			DispatchMessage(&msg);		// this will push it to the Window procedure
		}
	}
}

//////////////////////////////////////////////////////////
// Read and process the load files
//////////////////////////////////////////////////////////
void LoadOneImg(struct IMG *pImg, char *szFork) {
	FILE *fp;
	char *pData;
	HRSRC hRsrc;
	HGLOBAL hGlob;
	unsigned char DiskFile[0x10000+6];	// 64k ROM plus 6 byte kracker header max
	char *pszFrom="resource";

	if ((NULL == pImg) || (pImg->nType == TYPE_NONE)) return;

	pData=NULL;

	int nLen=pImg->nLength;

	if (NULL != pImg->dwImg) {
		hRsrc=FindResource(NULL, MAKEINTRESOURCE(pImg->dwImg), szFork);
		if (hRsrc) {
			int nRealLen=SizeofResource(NULL, hRsrc);
			if (nLen > nRealLen) nLen=nRealLen;

			hGlob=LoadResource(NULL, hRsrc);
			if (NULL != hGlob) {
				pData=(char*)LockResource(hGlob);
			}
		}
	} else {
		// It's a disk file. Worse, it may be a disk file with a header. 
		// But we may be able to determine that.
		if (strlen(pImg->szFileName) == 0) {
			return;
		}
		fp=fopen(pImg->szFileName, "rb");
		if (NULL == fp) {
			debug_write("Failed to load '%s'", pImg->szFileName);
			return;
		}
		pszFrom="disk";
		int nRealLen=0;
		while (!feof(fp)) {
			nRealLen+=fread(&DiskFile[nRealLen], 1, 0x10000+6-nRealLen, fp);
		}
		fclose(fp);

		if (nRealLen > 6) {
			// Check for 6 byte header - our simple check is if
			// byte 0 is 0x00 or 0xff, and bytes 4/5 contain the
			// load address, then strip the first six bytes
			if ((DiskFile[0]==0x00) || (DiskFile[0]==0xff)) {	// a flag byte?
				if (DiskFile[4]*256+DiskFile[5] == pImg->nLoadAddr) {
					debug_write("Removing header from %s", pImg->szFileName);
					nRealLen-=6;
					memmove(DiskFile, &DiskFile[6], nRealLen);
				}
			}
		}
		if (nLen != nRealLen) {
			debug_write("Warning: size mismatch on %s - expected >%04x bytes but found >%04x", pImg->szFileName, pImg->nLength, nRealLen);
		}
		if (nLen > nRealLen) nLen=nRealLen;
		pData=(char*)DiskFile;
	}

	if (NULL != pData) {
		// finally ;)
		debug_write("Loading file from %s: Type %c, Bank %d, Address 0x%04X, Length 0x%04X", pszFrom, pImg->nType, pImg->nBank, pImg->nLoadAddr, nLen);

		switch (pImg->nType) {
			case TYPE_GROM:
				if (pImg->nLoadAddr+nLen > 65536) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=65536-pImg->nLoadAddr;
				}
				if (pImg->nBank > 0) {
					grombanking=1;
				}
				if (pImg->nBank == -1) {
					// copy to all banks
					for (int idx=0; idx<PCODEGROMBASE; idx++) {
						memcpy(&GROMBase[idx].GROM[pImg->nLoadAddr], pData, nLen);
					}
				} else {
					if ((pImg->nBank < PCODEGROMBASE) && (pImg->nBank >= 0)) {
						memcpy(&GROMBase[pImg->nBank].GROM[pImg->nLoadAddr], pData, nLen);
					} else {
						debug_write("Not loading to unsupported GROM bank %d", pImg->nBank);
					}
				}
				break;

			case TYPE_ROM:
				if (pImg->nLoadAddr+nLen > 65536) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=65536-pImg->nLoadAddr;
				}
				//memcpy(&CPU[pImg->nLoadAddr], pData, nLen);
				WriteMemoryBlock(pImg->nLoadAddr, pData, nLen);
				memset(&ROMMAP[pImg->nLoadAddr], 1, nLen);
				break;

			case TYPE_SPEECH:
				if (pImg->nLoadAddr+nLen > 65536) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=65536-pImg->nLoadAddr;
				}
				memcpy(&SPEECH[pImg->nLoadAddr], pData, nLen);
				break;

			case TYPE_XB:
				if (pImg->nLoadAddr-0x6000+nLen > 8192) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=8192-(pImg->nLoadAddr-0x6000);
				}
				memcpy(&CPU2[pImg->nLoadAddr-0x6000], pData, nLen);
				xb=1;		// xb bank loaded
				break;

			case TYPE_RAM:
				if (pImg->nLoadAddr+nLen > 65536) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=65536-pImg->nLoadAddr;
				}
				//memcpy(&CPU[pImg->nLoadAddr], pData, nLen);
				WriteMemoryBlock(pImg->nLoadAddr, pData, nLen);
				break;

			case TYPE_VDP:
				if (pImg->nLoadAddr+nLen > 16384) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=16384-pImg->nLoadAddr;
				}
				memcpy(&VDP[pImg->nLoadAddr], pData, nLen);
				break;

			case TYPE_DSR:	// always loads at >4000, the load address is the CRU base
				if (nLen > 8192) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=8192;
				}
				memcpy(&DSR[(pImg->nLoadAddr>>8)&0x0f][0], pData, nLen);
				break;

			case TYPE_DSR2:	// always loads at >4000, the load address is the CRU base
				if (nLen > 8192) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=8192;
				}
				memcpy(&DSR[(pImg->nLoadAddr>>8)&0x0f][0x2000], pData, nLen);
				break;

			case TYPE_PCODEG:
				if (pImg->nLoadAddr+nLen > 65536) {
					debug_write("%s overwrites memory block - truncating.", pImg->szFileName);
					nLen=65536-pImg->nLoadAddr;
				}
				memcpy(&GROMBase[PCODEGROMBASE].GROM[pImg->nLoadAddr], pData, nLen);
				break;

			case TYPE_AMS:
				// Force the AMS card to on (reset later will init it)
				sams_enabled=1;
				sams_size=3;
				// decode the RLE encoded data into the AMS memory block
				// We use both words as the size value to get a 32-bit size
				RestoreAMS((unsigned char*)pData, (pImg->nLoadAddr<<16)|nLen);
				break;

			case TYPE_KEYS:
				if (NULL != PasteString) {
					free(PasteString);
				}
				PasteString=(char*)malloc(nLen);
				memcpy(PasteString, pData, nLen);
				// make SURE it's terminated
				PasteString[nLen-1]='\0';
				PasteIndex=PasteString;
				break;

			default:
				break;
		}
	}
	
	// WIN32 does not require (or even permit!) us to unlock and release these objects
}

void readroms()
{ 
	int idx;

	// set memory banks to zero. Real TIs don't do this here
	memset(CPU, 0, 65536);
	memset(ROMMAP, 0, 65536);
	memset(CPU2, 0, 8192);
	memset(VDP, 0, 16384);
	memset(CRU, 1, 4096);
	CRU[0]=0;
	memset(DSR, 0, 16*16384);
	memset(key, 0, 256);

	pMagicDisk=NULL;
	pCurrentHelpMsg=NULL;

	for (idx=0; idx<=PCODEGROMBASE; idx++) {
		memset(GROMBase[idx].GROM, 0, sizeof(GROMBase[idx].GROM));
	}

	// Disable the PCode card (at >1F00)
	DSR[0xF][0]=0;
	xb=0;									// no XB bank loaded
	grombanking=0;							// not using grom banking
	nCurrentDSR=-1;							// no DSR paged in
	memset(nDSRBank, 0, sizeof(nDSRBank));	// not on second page of DSR
	bUseSDCard=false;						// not using SD by default
	SDG_Init(false);

	// load the always load files
	for (idx=0; idx<sizeof(AlwaysLoad)/sizeof(IMG); idx++) {
		LoadOneImg(&AlwaysLoad[idx], "ROMS");
	}

	// load the appropriate system ROMs - each system is a cart structure that contains 4 ROMs
	for (idx=0; idx<MAXROMSPERCART; idx++) {
		LoadOneImg(&Systems[nSystem].Img[idx], "ROMS");
	}

	// if there is a cart plugged in, load that too
	if (nCart != -1) {
		struct CARTS *pBank=NULL;

 		switch (nCartGroup) {
		case 0:	pBank=Apps; break;
		case 1: pBank=Games; break;
		case 2: pBank=Users; break;
		case 3: bUseSDCard=true; 
				debug_write("Init SDGROM...");
				SDG_Init(true);
				pCurrentHelpMsg="SDGROM files are located in the SDGROM subfolder";
				break;
		}

		if (pBank) {
			pCurrentHelpMsg=pBank[nCart].szMessage;
			for (idx=0; idx<MAXROMSPERCART; idx++) {
				LoadOneImg(&pBank[nCart].Img[idx], "ROMS");
			}
			pMagicDisk=pBank[nCart].pDisk;
		}
	}
}

//////////////////////////////////////////////////////////
// Convert hex to decimal
//////////////////////////////////////////////////////////
Word hex2dec(char *x)
{
	Word z;

	z=0;
	sscanf(x, "%4hx", &z);

	return z;
}

//////////////////////////////////////////////////////////
// Speedup routine - convert from TI floating point
//////////////////////////////////////////////////////////
double tifloat2double(Byte *p) 
{
	short int firstword;
	int i, exp, sign=1;
	double res;
  
	firstword = *p++ << 8;
	firstword |= *p++;
	if (firstword==0)
	{
		return 0;
	}
	else 
	{
		if (firstword<0) 
		{								/* negative */
			sign=-1;
			firstword = -firstword;
		}
	}
	exp = ((firstword >> 8) - 0x40 - 6) << 1;
	res = firstword & 0xff;
	for (i=0; i<6; i++)
	{
		res = 100*res + *p++;
	}
	
	return (sign * res * (double)pow((double)10.0,exp));
}
	
//////////////////////////////////////////////////////////
// Speedup routine - convert to TI floating point and set status+error flags
//////////////////////////////////////////////////////////
void double2tifloat(double d, Byte *p)
{
	short int firstword;
	int i, exp, sign=1;
  
	wcpubyte(0x8354, 0);		/* reset error flag */
	if (d==0)
	{
		*p++ = 0;
		*p = 0;
		wcpubyte(0x8354, 0x20);	/* zero flag in GPL status */
		return;
	} 
	else
	{
		if (d<0) 
		{						/* negative */
			sign=-1;
			d = -d;
		}
	}
	exp = (int) log10(d) >> 1;
	d /= pow(10.0,exp<<1);
	d += 0.0000000000005;		/* round last digit */
	if (d>=100) {				/* and handle overflow here... */
	  d /= 100;
	  exp++;
	}
	if (exp<=-128) {			/* underflow */
	  *p++ = 0;
	  *p = 0;
	  wcpubyte(0x8354, 0x20);	/* zero flag in GPL status */
	  return;
	} else if (exp>=128) {		/* overflow */
	  exp = 127;
	  d = 99;
	  wcpubyte(0x8354, 0x01);	/* set error flag (FIXME: 02 for division) */
	}
	firstword = (exp + 0x40) << 8 | (Byte) d;
	if (sign<0) {
	  firstword = -firstword;
	  wcpubyte(0x837c, 0x00);	/* GPL status: reset Arithmetic Greater Than */
	} else {
	  wcpubyte(0x837c, 0x40);	/* GPL status: set Arithmetic Greater Than */
	}
	*p++ = firstword >> 8;
	*p++ = firstword & 0xff;
	for (i=0; i<6; i++) 
	  {
		d = (d - (int) d) * 100;
		*p++ = (Byte) d;
	  }
}

//////////////////////////////////////////////////////////
// Interpret a single instruction
//////////////////////////////////////////////////////////
void do1()
{
	cpucount++;										// increment instruction counter

	// check for video interrupt or redraw
	if (interrupt_needed)
	{
		nCycleCount=0;									// for throttling

		while (retrace_count>drawspeed)
		{
			if (redraw_needed)	// only draw if VDP RAM has been changed, and it's time
			{ 
				if (DisplayEvent) {
					SetEvent(DisplayEvent);
				}

				redraw_needed=0;
			}

			cpuframes++;								// for accounting
			retrace_count-=drawspeed+1;					// reset retrace counter
			timercount+=drawspeed+1;					// update timer statistics

			if (Recording)								// for AVI recording
			{
				// write a frame to the AVI
				// I want only every 4th frame.
				RecordFrame++;
				if (RecordFrame >= 4/(drawspeed+1))
				{
					RecordFrame=0;
					WriteFrame();
				}
			}
		}

		VDPS|=0x80;									// set interrupt flag
		interrupt_needed=0;							// No matter what, this tick is passed!
	}

	// Somewhat better 9901 timing - getting close! (Actually this looks pretty good now)
	// The 9901 timer decrements every 64 periods, with the TI having a 3MHz clock speed
	// Thus, it decrements 46875 times per second. If CRU bit 3 is on, we trigger an
	// interrupt when it expires. 
	if ((timer9901>0)&&(CRU[0] == 0)) {
		int nTimerCnt=CRUTimerTicks>>6;		// /64
		if (nTimerCnt) {
			timer9901-=nTimerCnt;
			CRUTimerTicks-=(nTimerCnt<<6);	// *64

			if (timer9901 < 1) {
				timer9901=starttimer9901;
				if (timer9901IntOk) {
//					debug_write("9901 timer expired, int requested");
					timer9901IntReq=1;
					timer9901IntOk=0;
//				} else {
//					debug_write("9901 timer expired, int not valid");
				}
			}
		}
	}

	// Check if the VDP or CRU wants an interrupt (99/4A has only level 1 interrupts)
	// When we have peripheral card interrupts, they are masked on CRU[1]
	if ((((VDPINT)&&(CRU[2]))||((timer9901IntReq)&&(CRU[3]))) && ((ST&0x000f) >= 1) && (!skip_interrupt)) {
		idling=0;								// stop IDLE instruction
		if (cycles_left >= 1) {					// speed throttling
#ifdef ALLOW_OLD_HACKS
			if (simulate_int) {
				if (timer9901IntReq) {
					timer9901IntReq=0;			// not supported in simulated mode
				} else {
					ConsoleInterrupt();				// Call console interrupt routine
				}
			} else 
#endif
			{
				wrword(0x83da,WP);				// WP in R13 
				wrword(0x83dc,PC);				// PC in R14 
				wrword(0x83de,ST);				// ST in R15 
				//ST=(ST&0xfff0);				// disable interrupts
				
				/* now load the correct workspace, and perform a branch and link to the address */
				/* (level 1 interrupt assumed as that's all we have) */
				WP=romword(0x0004);
				PC=romword(0x0006);

				nCycleCount+=22;
			}
		}
	}

	// Control keys - active only with scroll lock in PS/2 mode
	if ((keyboard != KEY_994A_PS2) || (0 != scrolllock)) {
		if (key[VK_HOME]) 
		{
			if (NULL == dbgWnd) {
				dbgWnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_DEBUG), myWnd, DebugBoxProc);
				ShowWindow(dbgWnd, SW_SHOW);
				// the dialog focus switch will cause a loss of the up event, so just fake it now
				decode(0xe0);	// extended key
				decode(0xf0);
				decode(VK_HOME);
			}
			key[VK_HOME]=0;
		}

		if (key[VK_END]) {
			if (dbgWnd) {
				SendMessage(dbgWnd, WM_CLOSE, 0, 0);
				decode(0xe0);	// extended key
				decode(0xf0);	// just in case
				decode(VK_END);
			}
			key[VK_END]=0;
		}

		if (key[VK_PRIOR]) {
			VDPDebug=1;
			key[VK_PRIOR]=0;
			redraw_needed=1;
		}
		if (key[VK_NEXT]) {
			VDPDebug=0;
			key[VK_NEXT]=0;
			redraw_needed=1;
		}

		// breakpoint handling
		if (max_cpf > 0) {
			// used for timing
			static unsigned long nFirst=0;
			static unsigned long nMax=0, nMin=0xffffffff;
			static int nCount=0;
			static unsigned long nTotal=0;

			for (int idx=0; idx<nBreakPoints; idx++) {
				switch (BreakPointType[idx]) {
					case BREAK_PC:
						if (PC == BreakPoint[idx]) {
							TriggerBreakPoint();
						}
						break;

					// timing instead of breakpoints
					case BREAK_START_TIMER:
						if (PC == BreakPoint[idx]) {
							nFirst=total_cycles;
						}
						break;

					case BREAK_EXIT_TIMER:
						if (PC == BreakPoint[idx]) {
							if (nFirst!=0) {
								if (total_cycles<nFirst) {
									debug_write("Counter Wrapped, no statistics");
								} else {
									unsigned long nTime=total_cycles-nFirst;
									if (nTime > 0) {
										if (nTime>nMax) nMax=nTime;
										if (nTime<nMin) nMin=nTime;
										nCount++;
										nTotal+=nTime;
										if (nTotal<nTime) {
											nTotal=0;
											nCount=0;
										} else {
											debug_write("Timer: %u CPU cycles - Min: %u  Max: %u  Average(%u): %u", nTime, nMin, nMax, nCount, nTotal/nCount);
										}
									}
								}
								nFirst=0;
							}
						}
						break;
				}
			}
		}

		// pause/play
		if (key[VK_F1]) {	
			if (0 == max_cpf) {
				// already paused, restore
				DoPlay();
			} else {
				// running normal, so pause
				DoPause();
			}
			key[VK_F1]=0;
		}
		
		// step
		if (key[VK_F2]) {
			DoStep();
			key[VK_F2]=0;
		}				
		// toggle turbo
		if (key[VK_F3]) {
			if (CPUThrottle==CPU_MAXIMUM) {
				// running in fast forward, return to normal
				DoPlay();
			} else {
				DoFastForward();
			}
			key[VK_F3]=0;
		}									
	}

	////////////////// System Patches

	// If an idle is set
	if ((idling)||(CPUSpeechHalt)) {
		cycles_left=0;
		if (CPUSpeechHalt) {
			SpeechUpdate(-1);		// try to remove the backlog
		}
	}
	if (cycles_left > 0)		// speed throttling
	{
		// Keyboard hacks - requires original 99/4A ROM!
		if ((nSystem == 1) && (PC==0x478)) {	// This is where KSCAN begins to exit
			if (NULL != PasteString) {
				
				if (PasteString==PasteIndex) SetWindowText(myWnd, "Classic99 - Pasting (ESC Cancels)");	

				if (key[VK_ESCAPE]) {
					*PasteIndex='\0';
				}

				if ((rcpubyte(0x8374)==0)||(rcpubyte(0x8374)==5)) {		// Check for pastestring - note keyboard is still active
					if (*PasteIndex) {
						if (*PasteIndex==10) {
							// CRLF to CR, LF to CR
							if ((PasteIndex==PasteString)||(*(PasteIndex-1)!=13)) {
								*PasteIndex=13;
							}
						}

						if (PasteCount<1) {
							if ((*PasteIndex>31)||(*PasteIndex==13)) {
								wcpubyte(WP, *PasteIndex);					/* set R0 (byte) with keycode */
								wcpubyte(WP+12, rcpubyte(0x837c) | 0x20);	/* R6 must contain the status byte */
							}
							if (PasteCount<1) wcpubyte(0x837c, rcpubyte(0x837c)|0x20);
							PasteCount++;
						} else {
							PasteCount++;
							if (PasteCount==2) {
								PasteIndex++;
							}
							if (PasteCount==3) {
								PasteCount=0;
							}
						}

					} else {
						PasteIndex=NULL;
						free(PasteString);
						PasteString=NULL;
						SetWindowText(myWnd, "Classic99");
					}
				}
			}
		}

		// intercept patches for Classic99 DSK1 emulation
		// These patches handle the custom DSR:
		// >4800 - DSK file access
		// >4810 - DSK powerup
		// >4820->482C - SBRLNK - Special subprograms
		if (nCurrentDSR == 1) {
			switch (PC) {
			case 0x4800:
				do_dsrlnk();
				PC=romword(WP+22)+2;						// return address in R11.. TI is a bit silly
															// and if we don't increment by 2, it's considered
															// an error condition
				return;

			case 0x4810:
				wrword(0x8358, 0x3ef5);						// volume ID block
				wrword(0x8366, 0x3eea);						// VDP stack
				do_files(3);
				PC=romword(WP+22);
				return;

			case 0x4820:
			case 0x4821:
			case 0x4822:
			case 0x4823:
			case 0x4824:
			case 0x4825:
			case 0x4826:
			case 0x4827:
			case 0x4828:
			case 0x4829:
			case 0x482a:
			case 0x482b:
			case 0x482c:
				do_sbrlnk();
				PC=romword(WP+22)+2;						// return address in R11.. as above
				return;
			}
		}

#ifdef ALLOW_OLD_HACKS
		if (speedup_float) 
		{									/* speedup floating point operations */
			if (PC==0x0d80) {					/* ADD */
				double2tifloat(tifloat2double(&CPU[0x835c]) + tifloat2double(&CPU[0x834a]), &CPU[0x834a]);
				PC=romword(WP+22);			// return address in R11.
				return;
			}
			if (PC==0x0d7c) {					/* SUBT */
				double2tifloat(tifloat2double(&CPU[0x835c]) - tifloat2double(&CPU[0x834a]), &CPU[0x834a]);
				PC=romword(WP+22);			// return address in R11.
				return;
			}
			if (PC==0xe88) {					/* MULT */
				double2tifloat(tifloat2double(&CPU[0x835c]) * tifloat2double(&CPU[0x834a]), &CPU[0x834a]);
				PC=romword(WP+22);			// return address in R11.
				return;
			}
			if (PC==0x0ff4) {					/* DIV */
				double den = tifloat2double(&CPU[0x834a]);
				if (den) {
				  double2tifloat(tifloat2double(&CPU[0x835c]) / den, &CPU[0x834a]);
				} else {
				  double2tifloat((CPU[0x835c] >= 0x80) ? -99e127 : 99e-127, &CPU[0x834a]);
				  //CPU[0x8354] = 0x02; /* set error */
				  WriteMemoryByte(0x8354, 0x02, false); /* set error */
				}
				PC=romword(WP+22);			// return address in R11.
				return;
			}
			if (PC==0x12b8) {					/* CFI */
				wrword(0x834a, (Word) tifloat2double(&CPU[0x834a]));
				PC=romword(WP+22);			// return address in R11.
				return;
			}
		} /* speedup_float */
#endif // SUPPRESS_FLOAT_CODE

		if (X_flag==0)
		{ 
			// Update the disassembly trace (warning: doesn't work with SDGROM)
			memmove(&disasm[0], &disasm[1], 38);//19*sizeof(Word));	// really should be a ring buffer
			disasm[19]=PC;

			in=romword(PC);							// ie: not an 'X' command
		    ADDPC(2);								// thanks to Jeff Brown for explaining that!
		}

		(*opcode[in])();							// 'in' contains the opcode to execute

		// Update the Software DAC pointer based on the CycleCount and an 8Khz audio output
		// 
		nDACCountDown-=nCycleCount;
		if (nDACCountDown < 1) {
			if (nDACPos < 500) {
				nDACPos++;
			} else {
				debug_write("DAC Overrun");
			}
			do {
				nDACCountDown+=DACUPDATECYCLECOUNT;
			} while (nDACCountDown<1);
		}

		// Slow down autorepeat using a timer - requires 99/4A GROM
		// We check if the opcode was "MOVB 2,*3+", (PC >025e, but we don't assume that) 
		// the 99/4A keyboard is on, and the GROM Address
		// has been incremented to the next instruction (IE: we just incremented the
		// repeat counter). If so, we only allow the increment at a much slower rate
		// based on the interrupt timer (for real time slowdown).
		if ((CPUThrottle!=CPU_NORMAL) && (slowdown_keyboard) && (in == 0xdcc2) && ((keyboard==KEY_994A)||(keyboard==KEY_994A_PS2)) && (GROMBase[0].GRMADD == 0x2a62)) {
			if ((ticks%10) != 0) {
				//CPU[0x830D]--;
				WriteMemoryByte(0x830D, ReadMemoryByte(0x830D) - 1, false);
			}
		}

		InterlockedExchangeAdd((LONG*)&cycles_left, -nCycleCount);
		unsigned long old=total_cycles;
		InterlockedExchangeAdd((LONG*)&total_cycles, nCycleCount);
		if ((old&0x80000000)&&(!(total_cycles&0x80000000))) {
			total_cycles_looped=true;
			speech_cycles=total_cycles;
		} else {
			// check speech
			if (max_cpf > 0) {
				if (total_cycles - speech_cycles >= (unsigned)SPEECHUPDATECOUNT) {
					while (total_cycles - speech_cycles >= (unsigned)SPEECHUPDATECOUNT) {
						static int nCnt=0;
						int nSamples=26;
						// should only be once
						// now we're expecting 1/300th of a second, which at 8khz
						// is 26.6 samples
						nCnt++;
						if (nCnt > 2) {	// handle 2/3
							nCnt=0;
						} else {
							nSamples++;
						} 
						SpeechUpdate(nSamples);
						speech_cycles+=SPEECHUPDATECOUNT;
					}
				}
			}
		}

		CRUTimerTicks+=nCycleCount;
		nCycleCount=0;

		if (bDebugAfterStep) {
			bDebugAfterStep=false;
			draw_debug();
		}
	} else {
		// Go to sleep till it's time to work again, timeout wait after 50ms
		// (so if we're very slow or the event dies, we keep running anyway)
		if (bDebugAfterStep) {
			bDebugAfterStep=false;
			draw_debug();
		}

		WaitForSingleObject(hWakeupEvent, 50);
	}

	skip_interrupt=0;
}

//////////////////////////////////////////////////////
// Read a single byte from CPU memory
//////////////////////////////////////////////////////
Byte rcpubyte(Word x,bool rmw) {
	// TI CPU memory map
	// >0000 - >1fff  Console ROM
	// >2000 - >3fff  Low bank RAM
	// >4000 - >5fff  DSR ROMs
	// >6000 - >7fff  Cartridge expansion
	// >8000 - >9fff  Memory-mapped devices & CPU scratchpad
	// >a000 - >ffff  High bank RAM
	// 
	// All is fine and dandy, except the memory-mapped devices, and the
	// fact that writing to the cartridge port with XB in places causes
	// a bank switch. In this emulator, that will only happen if bank 2
	// has been loaded. A custom DSR will be written to be loaded which
	// will appear to the "TI" to support all valid devices, and it will
	// be loaded into the DSR space.

	if (!rmw) {
		// Check for read or access breakpoints
		for (int idx=0; idx<nBreakPoints; idx++) {
			switch (BreakPointType[idx]) {
				case BREAK_ACCESS:
				case BREAK_READ:
					if (x == BreakPoint[idx]) {
						TriggerBreakPoint();
					}
					break;
			}
		}
	}

	nCycleCount+=2;	// 1/2 of the word access - this will be wrong for the places that directly read bytes!

	switch (x & 0xe000) {
		case 0x8000:
			switch (x & 0xfc00) {
				case 0x8000:				// scratchpad RAM - 256 bytes repeating.
					nCycleCount-=2;			// never mind for scratchpad :)
					return ReadMemoryByte(x | 0x0300);	// I map it all to >83xx
				case 0x8400:				// Don't read the sound chip (can hang a real TI)
					return 0;
				case 0x8800:				// VDP read data
					return(rvdpbyte(x));
				case 0x8c00:				// VDP write data
					return 0;
				case 0x9000:				// Speech read data
					return(rspeechbyte(x));
				case 0x9400:				// Speech write data
					return 0;
				case 0x9800:				// read GROM data
					{
						// always access console GROMs to keep them in sync
						Byte nRet = rgrmbyte(x);
						if (bUseSDCard) {
							Word nTmp=Classic99_SDGAccess(x, 0, GROM_ACCESS);
							if (0xffff != nTmp) {
								nRet=(nTmp&0xff);
							}
						}
						return nRet;
					}
				case 0x9c00:				// write GROM data
					return 0;
				default:					// We shouldn't get here, but just in case...
					return 0;
			}
		case 0x0000:					// console ROM
			nCycleCount-=2;				// never mind for console ROM :)
			// fall through
		case 0x2000:					// normal CPU RAM
		case 0xa000:					// normal CPU RAM
		case 0xc000:					// normal CPU RAM
		case 0xe000:					// normal CPU RAM
			return ReadMemoryByte(x);
		case 0x4000:					// DSR ROM (with bank switching and CRU)
			if (-1 == nCurrentDSR) return 0;

			// special case for P-Code Card GROM addresses
			if (nCurrentDSR == 0xf) {
				switch (x) {
					case 0x5bfc:		// read grom data
					case 0x5bfd:
					case 0x5bfe:		// read grom address
					case 0x5bff:
						return(rpcodebyte(x));

					case 0x5ffc:		// write data
					case 0x5ffd:
					case 0x5ffe:		// write address
					case 0x5fff:
						return 0;
				}
				// any other case, fall through
			}

			// SAMS support
			if (nCurrentDSR == 0xe) {
				bool read = ((x >= 0x4000) && (x < 0x4020));

				if (read && MapperRegistersEnabled()) {
					Byte reg = (x & 0xfe) >> 1;
					bool hiByte = ((x & 1) == 0);

					return ReadMapperRegisterByte(reg, hiByte);
				}
				return 0;
			}

			if (nDSRBank[nCurrentDSR]) {
				return DSR[nCurrentDSR][x-0x2000];	// page 1: -0x4000 for base, +0x2000 for second page
			} else {
				return DSR[nCurrentDSR][x-0x4000];	// page 0: -0x4000 for base address
			}
			break;

		case 0x6000:					// cartridge ROM
			// XB is supposed to only page the upper 4k, but some other carts seem to like it all
			// paged? Dunno.. weird.
			if (bUseSDCard) {
				Word nRet=Classic99_SDGAccess(x, 0, ROM_READ);
				if (0xffff == nRet) {
					if (xb && bank) {
						return(CPU2[x-0x6000]);	// cartridge bank 2
					} else {
						return ReadMemoryByte(x);	// cartridge bank 1
					}
				} else {
					return (Byte)nRet&0xff;
				}
			} else {
				if (xb && bank) {
					return(CPU2[x-0x6000]);	// cartridge bank 2
				} else {
					return ReadMemoryByte(x);	// cartridge bank 1
				}
			}
			break;

		default:						// We shouldn't get here, but just in case...
			return 0;
	}
}

//////////////////////////////////////////////////////////
// Write a byte to CPU memory
//////////////////////////////////////////////////////////
void wcpubyte(Word x, Byte c)
{
	// Check for write or access breakpoints
	for (int idx=0; idx<nBreakPoints; idx++) {
		switch (BreakPointType[idx]) {
			case BREAK_ACCESS:
			case BREAK_WRITE:
				if (x == BreakPoint[idx]) {
					TriggerBreakPoint();
				}
				break;
		}
	}

	nCycleCount+=2;	// 1/2 of the word access time

	if ((x>=0x6000)&&(x<0x8000)) {
		if (bUseSDCard) {
			Classic99_SDGAccess(x, c, ROM_WRITE);
			goto checkmem;
		} else {
			if ((xb) && (ROMMAP[x])) {		// trap ROM writes and check for XB bank switch
				bank=(x&0x0002)>>1;			// XB bank switch
				goto checkmem;
			}
			// else it's RAM there
		}
	}

	switch (x & 0xe000) {
		case 0x8000:
			switch (x & 0xfc00) {
				case 0x8000:				// scratchpad RAM - 256 bytes repeating.
				nCycleCount-=2;				// never mind for scratchpad
				WriteMemoryByte((x | 0x0300), c, false);  // I map it all to >83xx
				break;
			case 0x8400:				// Sound write data
				wsndbyte(c);
				break;
			case 0x8800:				// VDP read data
				break;
			case 0x8c00:				// VDP write data
				wvdpbyte(x,c);
				break;
			case 0x9000:				// Speech read data
				break;
			case 0x9400:				// Speech write data
				wspeechbyte(x,c);
				break;
			case 0x9800:				// read GROM data
				break;
			case 0x9c00:				// write GROM data
				if (bUseSDCard) {
					Classic99_SDGAccess(x, c, GROM_ACCESS);
					// need to keep the console GROMs in sync, so don't return
				}
				wgrmbyte(x,c);
				break;
			default:					// We shouldn't get here, but just in case...
				break;
		}
		break;

		case 0x0000:					// console ROM
			nCycleCount-=2;				// never mind for ROM
			// fall through
		case 0x2000:					// normal CPU RAM
		case 0x6000:					// Cartridge RAM (ROM is trapped above)
		case 0xa000:					// normal CPU RAM
		case 0xc000:					// normal CPU RAM
		case 0xe000:					// normal CPU RAM
			WriteMemoryByte(x, c, false);
			break;
		case 0x4000:					// DSR ROM - read only for now
			// SAMS support
			if (nCurrentDSR == 0xe) {
				bool change = ((x >= 0x4000) && (x < 0x4020));

				if (change && MapperRegistersEnabled()) {
					Byte reg = (x & 0xfe) >> 1;
					bool hiByte = ((x & 1) == 0);

					WriteMapperRegisterByte(reg, c, hiByte);
				}
				return;
			}

			// special case for P-Code Card GROM addresses
			if (nCurrentDSR == 0xf) {
				switch (x) {
					case 0x5bfc:		// read grom data
					case 0x5bfd:
					case 0x5bfe:		// read grom address
					case 0x5bff:
						break;

					case 0x5ffc:		// write grom data
					case 0x5ffd:
					case 0x5ffe:		// write grom address
					case 0x5fff:
  						wpcodebyte(x,c);
						goto checkmem;
				}
				// any other case, fall through, but we currently treat all DSR memory as ROM
			}
  			break;
	}

checkmem:
	// check breakpoints against what was written to where
	for (int idx=0; idx<nBreakPoints; idx++) {
		switch (BreakPointType[idx]&0xff) {
			case BREAK_EQUALS_BYTE:
				if ((x == BreakPoint[idx]) && (c == ((BreakPointType[idx]&0xff00)>>8))) {
					TriggerBreakPoint();
				}
				break;
		}
	}
}

//////////////////////////////////////////////////////
// Read a byte from the speech synthesizer
//////////////////////////////////////////////////////
Byte rspeechbyte(Word x)
{
	Byte ret=0;

	if (SpeechRead) {
		ret=SpeechRead();
	}
	return ret;
}

//////////////////////////////////////////////////////
// Write a byte to the speech synthesizer
//////////////////////////////////////////////////////
void wspeechbyte(Word x, Byte c)
{
	if ((SpeechWrite)&&(SpeechEnabled)) {
		if (!SpeechWrite(c, false)) {
			debug_write("Speech halt triggered.");
			CPUSpeechHalt=true;
			CPUSpeechHaltByte=c;
		}
	}
}

//////////////////////////////////////////////////////
// Speech Update function - runs every x instructions
// Pass in number of samples to process.
//////////////////////////////////////////////////////
void SpeechUpdate(int nSamples) {
	if ((speechbuf==NULL) || (SpeechProcess == NULL)) {
		CPUSpeechHalt=false;
		return;
	}

	if (nSamples == -1) {
		if (CPUSpeechHalt) {
			// we want to clear the backlog, no more than 200 samples (1 frame)
			int nCountDown=20;
			while ((nCountDown--) && (!SpeechWrite(CPUSpeechHaltByte, true))) {
				if (nSpeechTmpPos+10 >= SPEECHRATE*2) {
					// in theory, this should not happen, though it may if we
					// run the 9900 very fast
					debug_write("Speech buffer full... dropping");
					return;
				}

				SpeechProcess((unsigned char*)&SpeechTmp[nSpeechTmpPos], 10);
				nSpeechTmpPos+=10;
				Sleep(1);
			}
			if (nCountDown) {
				debug_write("Speech halt cleared");
			} else {
				debug_write("Speech halt timed out, continuing.");
			}
			CPUSpeechHalt=false;
		} else {
			debug_write("Called speech_update without sample count and CPU not halted");
			return;
		}
	} else {
		if (nSpeechTmpPos+nSamples >= SPEECHRATE*2) {
			// in theory, this should not happen, though it may if we
			// run the 9900 very fast
			debug_write("Speech buffer full... dropping");
			return;
		}

		SpeechProcess((unsigned char*)&SpeechTmp[nSpeechTmpPos], nSamples);
	}

	// copy the DAC buffer into the speech buffer (this is why it's so much smaller)
	// DAC sound will overwrite speech but silence won't. Values are from 0 (silent) to 15 (loudest)
	// We need to convert to signed 16-bit, we'll use positive values. hopefully this never
	// gets ahead of speech
	for (int idx=0; idx<nDACPos; idx++) {
		if (DACTmp[idx] > 0) {
			SpeechTmp[nSpeechTmpPos+idx]=DACTmp[idx];	// should be a DAC volume adjust
		}
	}
	memset(DACTmp, 0, sizeof(DACTmp));

	if (nDACPos>nSamples+10) {	// +10 to allow a little slack
		if (max_cpf == DEFAULT_CPF) {
			// only warn when at 100% speed
			debug_write("DAC samples ahead of Speech buffer");
		}
	}

	nDACPos=0;
	nSpeechTmpPos+=nSamples;
}

void SpeechBufferCopy() {
	DWORD iRead, iWrite;
	Byte *ptr1, *ptr2;
	DWORD len1, len2;
	static DWORD lastRead=0;

	if (nSpeechTmpPos == 0) {
		// no data to write
		return;
	}

	// just for statistics
	speechbuf->GetCurrentPosition(&iRead, &iWrite);
//	debug_write("Read/Write bytes: %5d/%5d", iRead-lastRead, nSpeechTmpPos*2);
	lastRead=iRead;

	if (SUCCEEDED(speechbuf->Lock(iWrite, nSpeechTmpPos*2, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_FROMWRITECURSOR))) 
	{
		memcpy(ptr1, SpeechTmp, len1);
		if (len2 > 0) {							// handle wraparound
			memcpy(ptr2, &SpeechTmp[len1/2], len2);
		}
		speechbuf->Unlock(ptr1, len1, ptr2, len2);	
		
		// reset the buffer
		nSpeechTmpPos=0;
	} else {
		debug_write("Speech buffer lock failed\n");
		// don't reset the buffer, we may get it next time
	}
}

//////////////////////////////////////////////////////
// 'DAC' Update function - drops a brief sample of
// volume level into the speech buffer 
// running Speech and DAC at the same time won't
// work properly, but it seems unlikely to happen ;)
// Also, Barry Boone reports that his SoundFX plays
// back at up to 11kHz, but the speech buffer is only
// 8kHz, that may cause samples to sound worse than
// they should!!
//////////////////////////////////////////////////////
void DACUpdate(int val) {
	// this isn't quite right either cause we copy
	// a few bytes into the buffer then trail off,
	// ideally we should write the new value every 
	// time.
	// Also, multiple channel DAC will just overwrite
	// itself with the last channel set. That could
	// be an issue as well.
	DACTmp[nDACPos]=val;
	DACTmp[nDACPos+1]=val;
	DACTmp[nDACPos+2]=val;
	DACTmp[nDACPos+3]=val;
	DACTmp[nDACPos+4]=val;
	DACTmp[nDACPos+5]=val;
	DACTmp[nDACPos+6]=val;
	DACTmp[nDACPos+7]=val;
	DACTmp[nDACPos+8]=val;
	DACTmp[nDACPos+9]=val;
}

//////////////////////////////////////////////////////
// Increment VDP Address
//////////////////////////////////////////////////////
void increment_vdpadd() 
{
	VDPADD=(++VDPADD) & 0x3fff;
}

//////////////////////////////////////////////////////////////
// Read from VDP chip
//////////////////////////////////////////////////////////////
Byte rvdpbyte(Word x)
{ 
	unsigned short z;

	if ((x>=0x8c00) || (x&1))
	{
		return(0);											// write address
	}

	if (x&0x0002)
	{	/* read status */
		z=VDPS;
		VDPS&=0x1f;			// reset interrupt, five sprite and collision bits
		vdpaccess=0;		// reset byte flag
		return((Byte)z);
	}
	else
	{ /* read data */
		z=vdpprefetch;
		vdpprefetch=VDP[VDPADD];
		increment_vdpadd();
		vdpaccess=0;
		return ((Byte)z);
	}
}

///////////////////////////////////////////////////////////////
// Write to VDP chip
///////////////////////////////////////////////////////////////
void wvdpbyte(Word x, Byte c)
{
	if (x<0x8c00 || (x&1)) 
	{
		return;							/* not going to write at that block */
	}

	if (x&0x0002)
	{	/* write address */
		tmpVDPADD=(tmpVDPADD>>8)|(c<<8);
		vdpaccess++;
		if (vdpaccess==2)
		{ 
			if (tmpVDPADD&0x8000)
			{ 
				wVDPreg((Byte)((tmpVDPADD&0x0700)>>8),(Byte)(tmpVDPADD&0x00ff));
				redraw_needed=1;
			} else {
				VDPADD=tmpVDPADD;
				if ((VDPADD&0x4000)==0) {	// set for reading?
					vdpprefetch=VDP[VDPADD];
					increment_vdpadd();
				} else {
					VDPADD&=0x3fff;			// writing, just mask the bit off
				}
			}
			vdpaccess=0;
		}
	}
	else
	{	/* write data */
		VDP[VDPADD]=c;

		// check breakpoints against what was written to where
		for (int idx=0; idx<nBreakPoints; idx++) {
			switch (BreakPointType[idx]&0xff) {
				case BREAK_EQUALS_VDP:
					if ((VDPADD == BreakPoint[idx]) && (c == ((BreakPointType[idx]&0xff00)>>8))) {
						TriggerBreakPoint();
					}
					break;
			}
		}

		vdpprefetch=c;
		increment_vdpadd();
		redraw_needed=1;
		vdpaccess=0;
	}
}

////////////////////////////////////////////////////////////////
// Write to VDP Register
////////////////////////////////////////////////////////////////
void wVDPreg(Byte r, Byte v)
{ 
	int t;

	VDPREG[r]=v;

	// check breakpoints against what was written to where
	for (int idx=0; idx<nBreakPoints; idx++) {
		switch (BreakPointType[idx]&0xff) {
			case BREAK_EQUALS_VDPREG:
				if ((r == BreakPoint[idx]) && (v == ((BreakPointType[idx]&0xff00)>>8))) {
					TriggerBreakPoint();
				}
				break;
		}
	}

	if (r==7)
	{	/* color of screen, set color 0 (trans) to match */
		t=v&0xf;
		if (t) {
			TIPALETTE[0]=TIPALETTE[t];
		} else {
			TIPALETTE[0]=0;
		}
		redraw_needed=1;
	}

}

////////////////////////////////////////////////////////////////
// Write a byte to the sound chip
// Nice notes at http://www.smspower.org/maxim/docs/SN76489.txt
////////////////////////////////////////////////////////////////
bool SetVoiceVolume(int c, int *oldVol) {
	unsigned int x;
	bool fDACChanged=false;
	int chan=(c&0x60)>>5;

	x=(c&0x0f);
	if (x==0xf) {
		voice[chan]->SetVolume(DSBVOLUME_MIN);
		DAC[chan]=0;
		if (fVoiceSimulatedDAC[chan]) fDACChanged=true;
	} else {
		DAC[chan]=15-x;
		if (fVoiceSimulatedDAC[chan]) {
			fDACChanged=true;
		} else {
			voice[chan]->SetVolume(x*(volume_mult));
		}
		if ((!fSampleShouldLoop[chan])&&(oldVol[chan]==0x0f)) {
			DWORD stat;
			if ((SUCCEEDED(voice[chan]->GetStatus(&stat)))&&(stat&DSBSTATUS_PLAYING)) {
				voice[chan]->SetCurrentPosition(0);
			} else {
				voice[chan]->Play(0, 0, 0);
			}
		}
	}
	oldVol[chan]=x;

	return fDACChanged;
}

void wsndbyte(Byte c)
{
	unsigned int x, idx;								// temp variable
	static int noise_volume=-10000;						// volume of noise
	static int noise_type=0;							// type of noise
	static int freq3=22000;								// frequency calculator
	static int oldVol[3]={0,0,0};						// tone generator old volumes
	static int oldFreq[3]={0,0,0};						// tone generator frequencies
	bool fDACChanged=false;

#if 0
	static FILE *fLog=NULL;
	if (NULL == fLog) {
		fLog=fopen("C:\\AUDIO.LOG", "wb");
	}
	if (NULL != fLog) {
		int nIdx;
		switch (c&0xf0) {
			case 0x80:	nIdx=0; break;
			case 0x90:	nIdx=5; break;
			case 0xA0:	nIdx=10; break;
			case 0xB0:	nIdx=15; break;
			case 0xC0:	nIdx=20; break;
			case 0xD0:	nIdx=25; break;
			case 0xE0:	nIdx=30; break;
			case 0xF0:	nIdx=33; break;
			default:
				switch (latch_byte&0xf0) {
				case 0x80:	nIdx=2; break;
				case 0xA0:	nIdx=12; break;
				case 0xC0:	nIdx=22; break;
				default: nIdx=27; break;
				}
		}

		fprintf(fLog, "%*.*s%02X\n",nIdx,nIdx," ",c);

		//fputc(c, fLog);
	}

#endif

	if (NULL == lpds) return;

	// 'c' contains the byte currently being written to the sound chip
	// all functions are 1 or 2 bytes long, as follows					
	//
	// BYTE		BIT		PURPOSE											
	//	1		0		always '1' (latch bit)							
	//			1-3		Operation:	000 - tone 1 frequency				
	//								001 - tone 1 volume					
	//								010 - tone 2 frequency				
	//								011 - tone 2 volume					
	//								100 - tone 3 frequency				
	//								101 - tone 3 volume					
	//								110 - noise control					
	//								111 - noise volume					
	//			4-7		Least sig. frequency bits for tone, or volume	
	//					setting (0-F), or type of noise.				
	//					(volume F is off)								
	//					Noise set:	4 - always 0						
	//								5 - 0=periodic noise, 1=white noise 
	//								6-7 - shift rate from table, or 11	
	//									to get rate from voice 3.		
	//	2		0-1		Always '0'. This byte only used for frequency	
	//			2-7		Most sig. frequency bits						
	//
	// Commands are instantaneous

	// Latch anytime the high bit is set
	// This byte still immediately changes the channel
	if (c&0x80) {
		latch_byte=c;
	}

	switch (c&0xf0)										// check command
	{	
	case 0x90:											// Voice 1 vol
	case 0xb0:											// Voice 2 vol
	case 0xd0:											// Voice 3 vol
		fDACChanged=SetVoiceVolume(c, oldVol);
		break;

	case 0xe0:
		x=(c&0x07);										// Noise - get type
		for (idx=0; idx<8; idx++)
		{
			if (idx==x)
				noise[idx]->SetVolume(noise_volume);
			else
				noise[idx]->SetVolume(DSBVOLUME_MIN);	// turn off noises, except the one we want
		}
		noise_type=x;
		break;

	case 0xf0:											// Noise volume
		x=(c&0x0f);
		if (x==0xf) {
			noise[noise_type]->SetVolume(DSBVOLUME_MIN);
			noise_volume=DSBVOLUME_MIN;
		} else {
			noise[noise_type]->SetVolume(x*(volume_mult));
			noise_volume=x*(volume_mult);
		}
		break;

//	case 0x80:											// Voice 1 frequency
//	case 0xa0:											// Voice 2 frequency
//	case 0xc0:											// Voice 3 frequency
	default:											// Any other byte
		int nChan=(latch_byte&0x60)>>5;
		if (nChan > 2) break;
		if (c&0x80) {
			// least significant bits
			oldFreq[nChan]&=0xfff0;
			oldFreq[nChan]|=c&0x0f;
		} else {
			// most significant bits
			oldFreq[nChan]&=0xf;
			oldFreq[nChan]|=(c&0x3f)<<4;
		}
		x=oldFreq[nChan] + 1;			// Get frequency code (add 1 to prevent divide by zero)
		x=(111860/x)*22050;				// Make sample rate in HZ (22khz samples)
		//debug_write("Settting sample rate for voice %d to %dHZ", nChan, x/22050);
		if (x>220500000) {	// 10khz - more than that won't play well anyway! And x==0 is deliberate DAC mode
			if (false == fVoiceSimulatedDAC[nChan]) {
				fVoiceSimulatedDAC[nChan]=true;
//				debug_write("Enabling DAC mode on channel %d", nChan);
				fDACChanged=true;
			}
		} else {
			if (fVoiceSimulatedDAC[nChan]) {
				fVoiceSimulatedDAC[nChan]=false;
//				debug_write("Disabling DAC mode on channel %d", nChan);
				fDACChanged=true;
			}
		}
		x/=440;								// adjust to balance (sample recorded as 440hz)
		voice[nChan]->SetFrequency(x);		// set new frequency

		if (!fSampleShouldLoop[nChan]) {
			DWORD stat;
			if ((SUCCEEDED(voice[nChan]->GetStatus(&stat)))&&(stat&DSBSTATUS_PLAYING)) {
				voice[nChan]->SetCurrentPosition(0);	// ...and retrigger sample
			} else {
				voice[nChan]->Play(0, 0, 0);
			}
		}
		if (nChan==2)							// If third voice, save and adjust for noise channel
		{
			freq3=x;								
			noise[3]->SetFrequency(freq3);
			noise[7]->SetFrequency(freq3);
		}
		break;
	}

	if (fDACChanged) {
		// we need to update the DAC channel. For that, we cheat and use the speech synth
		// channel (although it being at only 8khz may be a problem, we'll see how it works)
		int val=0;	// signed 16-bit, so we'll use 0-0x7fff
		// Each channel contributes up to 10920 for a max of 32760 (with 32767 as max)
		// set the multiplier by the number of channels in use so we'll always get
		// maximum volume out of it.
		int nMult=728;
		if (!fVoiceSimulatedDAC[0]) nMult+=728;
		if (!fVoiceSimulatedDAC[1]) nMult+=728;
		if (!fVoiceSimulatedDAC[2]) nMult+=728;

		if (fVoiceSimulatedDAC[0]) val+=DAC[0]*nMult;
		if (fVoiceSimulatedDAC[1]) val+=DAC[1]*nMult;
		if (fVoiceSimulatedDAC[2]) val+=DAC[2]*nMult;
		DACUpdate(val);
	}
}

//////////////////////////////////////////////////////////////////
// GROM base 0 (console GROMS) manage all address operations
//////////////////////////////////////////////////////////////////
// Read a byte from GROM
//////////////////////////////////////////////////////////////////
Byte ReadValidGrom(int nBase, Word x) {
	Byte z;

	// the -1 accounts for the prefetch to get the data we're going to read
	if ((Word)(GROMBase[0].GRMADD-1) < 0x6000) {
		// console GROMs always respond
		nBase=0;
	}

//	if (nBase > 0) {
//		debug_write("Read GROM base %d(>%04X), >%04x, >%02x", nBase, x, GROMBase[0].GRMADD, GROMBase[nBase].grmdata);
//	}

	if (x&0x0002)
	{
		GROMBase[0].grmaccess=2;
		z=(GROMBase[0].GRMADD&0xff00)>>8;
		// read is destructive
		GROMBase[0].GRMADD=(((GROMBase[0].GRMADD&0xff)<<8)|(GROMBase[0].GRMADD&0xff));		
		return(z);
	}
	else
	{
		GROMBase[0].grmaccess=2;
		z=GROMBase[nBase].grmdata;
		// update all bases prefetch
		for (int idx=0; idx<PCODEGROMBASE; idx++) {
			GROMBase[idx].grmdata=GROMBase[idx].GROM[GROMBase[0].GRMADD];
		}
		GROMBase[0].GRMADD++;
		return(z);
	}
}

Byte rgrmbyte(Word x)
{
	unsigned int z;										// temp variable
	int nBank;

	if (x>=0x9c00)
	{
		return(0);										// write address
	}

	if (grombanking) {
		nBank=(x&0x3ff)>>2;								// maximum possible range to >9BFF - not all supported here though
		if (nBank >= PCODEGROMBASE) {
			debug_write("Invalid GROM base 0x%04X read", x);
			return 0;
		}
	} else {
		nBank=0;
	}

	return ReadValidGrom(nBank, x);
}

//////////////////////////////////////////////////////////////////
// Write a byte to GROM
//////////////////////////////////////////////////////////////////
void WriteValidGrom(int nBase, Word x, Byte c) {
//	if (nBase > 0) {
//		debug_write("Write GROM base %d(>%04X), >%04x, >%02x, %d", nBase, x, GROMBase[0].GRMADD, c, GROMBase[0].grmaccess);
//	}

	if (x&0x0002)
	{
		GROMBase[0].GRMADD=(GROMBase[0].GRMADD<<8)|(c);						// write GROM address
		GROMBase[0].grmaccess--;
		if (GROMBase[0].grmaccess==0)
		{ 
			GROMBase[0].grmaccess=2;										// prefetch emulation
			// update all bases prefetch
			for (int idx=0; idx<PCODEGROMBASE; idx++) {
				GROMBase[idx].grmdata=GROMBase[idx].GROM[GROMBase[0].GRMADD];
			}
			GROMBase[0].GRMADD++;
		}
	}
	else
	{
		GROMBase[0].grmaccess=2;
		if (nBase != PCODEGROMBASE) {
			debug_write("Writing to GROM!!");							// write to GROM not supported
		} else {
			debug_write("Writing to PCODE GROM!!");
		}
		//GROMBase[nBase].GROM[GROMBase[0].GRMADD]=c;		// don't write data (it's G*ROM*)
		GROMBase[0].GRMADD++;
	}
}


void wgrmbyte(Word x, Byte c)
{
	int nBank;

	if (x<0x9c00) 
	{
		return;											// read address
	}

	if (grombanking) {
		nBank=(x&0x3ff)>>2;								// maximum possible range to >9BFF - not all supported here though
		if (nBank >= PCODEGROMBASE) {
			debug_write("Invalid GROM base 0x%04X write", x);
			return;
		}
	} else {
		nBank=0;
	}

	return WriteValidGrom(nBank, x, c);
}

//////////////////////////////////////////////////////////////////
// Read a byte from P-Code GROM
//////////////////////////////////////////////////////////////////
Byte rpcodebyte(Word x)
{
	unsigned int z;										// temp variable

//	debug_write("Read PCODE GROM >%04x", GROMBase[PCODEGROMBASE].GRMADD);

	if (x>=0x5ffc)
	{
		return(0);										// write address
	}

	return ReadValidGrom(PCODEGROMBASE, x);
}

//////////////////////////////////////////////////////////////////
// Write a byte to P-Code GROM
//////////////////////////////////////////////////////////////////
void wpcodebyte(Word x, Byte c)
{
	if (x<0x5ffc) 
	{
		return;											// read address
	}

	WriteValidGrom(PCODEGROMBASE, x, c);
}

//////////////////////////////////////////////////////////////////
// Write a bit to CRU
//////////////////////////////////////////////////////////////////
void wcru(Word ad, int bt)
{
	static bool bTimerDirty=false;

	if (ad>=0x800) {
		// DSR space
		ad<<=1;		// put back into familiar space. A bit wasteful, but devices aren't high performance
		if (bt) {
			// bit 0 enables the DSR rom, so we'll check that first
			if ((ad&0xff) == 0) {
				nCurrentDSR=(ad>>8)&0xf;

				if (nCurrentDSR == 0x0e) {
					EnableMapperRegisters(true);
				}
//				debug_write("Enabling DSR at >%04x", ad);

				return;
			}
			// else, it's device dependent
			switch (ad&0xff00) {
			case 0x1e00:	// Memory card
				if ((ad & 0xff) == 0x02) {
					SetMemoryMapperMode(Map);
				}
				break;

			case 0x1f00:	// pCode card
				if ((ad&0xff) == 0x80) {
					// bank switch
//					debug_write("Switching P-Code to bank 2");
					nDSRBank[0xf]=1;
				}
				break;
			}
		} else {
			// bit 0 enables the DSR rom, so we'll check that first
			if ((ad&0xff) == 0) {
				if (((ad>>8)&0xf) == nCurrentDSR) {
					if (nCurrentDSR == 0x0e) {
						EnableMapperRegisters(false);
					}
					nCurrentDSR=-1;
//					debug_write("Disabling DSR at >%04x", ad);
					return;
				}
			}
			// else, it's device dependent
			switch (ad&0xff00) {
			case 0x1e00:	// memory card
				if ((ad & 0xff) == 0x02) {
					SetMemoryMapperMode(Passthrough);
				}
				break;

			case 0x1f00:	// pCode card
				if ((ad&0xff) == 0x80) {
					// bank switch
//					debug_write("Switching P-Code to bank 1");
					nDSRBank[0xf]=0;
				}
				break;
			}
		}
		return;
	} else {
		ad=(ad&0x0fff);										// get actual CRU line

//		debug_write("Write CRU 0x%x with %d", ad, bt);

		if (bt) {											// write the data
			if ((ad>0)&&(ad<16)&&(CRU[0]==1)) {
				if (ad == 15) {
					// writing 1 has no effect
				} else {
					// writing to CRU 9901 timer
					Word mask=0x01<<(ad-1);
					starttimer9901|=mask;
					bTimerDirty=true;
				}
			} else {
				CRU[ad]=1;
				switch (ad) {
					case 3:
						// timer interrupt bit
						timer9901IntReq=0;
						break;
					case 0x18:
						// ticking the cassette speaker
						DACUpdate(32000);	// high volume
						break;
				}
			}
		} else {
			if ((ad>0)&&(ad<16)&&(CRU[0]==1)) {
				if (ad == 15) {
					// writing 0 is a soft reset of the 9901 - it resets
					// all I/O pins to pure input, but does not affect the timer
					// Does it exit timer mode?
				} else {
					// writing to CRU 9901 timer
					Word mask=0x01<<(ad-1);
					starttimer9901&=~mask;
					bTimerDirty=true;
				}
			} else {
				CRU[ad]=0;
				switch (ad) {
					case 3:
						// timer interrupt bit
						timer9901IntReq=0;
						break;
					case 0:
						// Turning off timer mode - start timer
						if (bTimerDirty) {
							timer9901=starttimer9901;
							bTimerDirty=false;
						}
						CRUTimerTicks=0;
						timer9901IntOk=1;
	//					debug_write("Starting 9901 timer at %d ticks", timer9901);
						break;
					case 0x18:
						// ticking the cassette speaker
						DACUpdate(0);	// no volume
						break;
				}
			}
		}
		if ((ad > 15) && (ad < 31) && (CRU[0] == 1)) {
			// exit timer mode
			wcru(0,0);
		}
		// There's another potential case for automatic exit of timer mode
		// if a value from 16-31 appears on A10-A15 (remember A15 is LSB),
		// Thierry Nouspikel says that the 9901 will see this and exit 
		// timer mode as well, even though it's not a CRU operation.
		// That would mean any address access to >xx1x, >xx5x, >xx9x, >xxdx
		// should trigger it. Maybe we can test this on a real machine
		// sometime? TODO.
	}
}

//////////////////////////////////////////////////////////////////
// Read a bit from CRU
//////////////////////////////////////////////////////////////////
int CheckJoysticks(Word ad, int col) {
	int joyX, joyY, joyFire, joykey;
	int joy1col, joy2col;
	int ret=1;

	// Read external hardware
	joyX=0;
	joyY=0;
	joyFire=0;

	if (nSystem != 0) {
		// 99/4A
		joy1col=4;
		joy2col=0;
		joykey=1;
	} else {
		// 99/4
		joy1col=2;
		joy2col=4;
		joykey=0;
	}

	if ((col == joy1col) || (col == joy2col))				// reading joystick
	{	
		// TODO: This still reads the joystick many times for a single scan, but it's better ;)
		if (fJoy) {
			int device;

			device=-1;

			if (col==joy2col) {
				switch (joy2mode) {
					case 1: device=JOYSTICKID1; break;
					case 2: device=JOYSTICKID2; break;
				}
			} else {
				switch (joy1mode) {
					case 1: device=JOYSTICKID1; break;
					case 2: device=JOYSTICKID2; break;
				}
			}

			if (device!=-1) {
				memset(&myJoy, 0, sizeof(myJoy));
				myJoy.dwSize=sizeof(myJoy);
				myJoy.dwFlags=JOY_RETURNBUTTONS | JOY_RETURNX | JOY_RETURNY | JOY_USEDEADZONE;
				if (JOYERR_NOERROR == joyGetPosEx(device, &myJoy)) {
					if (0!=myJoy.dwButtons) {
						joyFire=1;
					}
					if (myJoy.dwXpos<0x4000) {
						joyX=-4;
					}
					if (myJoy.dwXpos>0xC000) {
						joyX=4;
					}
					if (myJoy.dwYpos<0x4000) {
						joyY=4;
					}
					if (myJoy.dwYpos>0xC000) {
						joyY=-4;
					}
				}
			} else {	// read the keyboard
				if (key[VK_TAB]) {
					joyFire=1;
				}
				if (key[VK_LEFT]) {
					joyX=-4;
				}
				if (key[VK_RIGHT]) {
					joyX=4;
				}
				if (key[VK_UP]) {
					joyY=4;
				}
				if (key[VK_DOWN]) {
					joyY=-4;
				}
			}
		}

		if (ad == 3)
		{	
			if ((key[KEYS[joykey][col][0]])||(joyFire))	// button reads normally
			{
				ret=0;
			}
		}
		else
		{
			if (key[KEYS[joykey][col][ad-3]])		// stick return (*not* inverted. Duh)
			{	
				ret=0;
			}
			if (ret) {
				switch (ad-3) {						// Check real joystick
				case 1: if (joyX ==-4) ret=0; break;
				case 2: if (joyX == 4) ret=0; break;
				case 3: if (joyY ==-4) ret=0; break;
				case 4: if (joyY == 4) ret=0; break;
				}
			}
		}
	}
	return ret;
}

int rcru(Word ad)
{
	int ret,col;										// temp variables

	if ((CRU[0]==1)&&(ad<16)&&(ad>0)) {				// read elapsed time from timer
		if (ad == 15) {
			return timer9901IntReq;
		}
//		if (ad==2) {
//			debug_write("Reading 9901 timer at %d ticks", timer9901);
//		}
		Word mask=0x01<<(ad-1);
		if (timer9901 & mask) {
			return 1;
		} else {
			return 0;
		}
	}

	if ((ad > 15) && (ad < 31) && (CRU[0] == 1)) {
		// exit timer mode
		wcru(0,0);
	}

	// only certain CRU bits are implemented, check them
	if (ad >= 0x1000) {
		// Plugin card - none supported yet
		return 1;	// false
	}

	// The CRU bits >0000 through >001f are repeated through the whole 4k range!
	ad=(ad&0x001f);										// get actual CRU line
	ret=1;												// default return code (false)

	// are we checking VDP interrupt?
	if (ad == 0x02) {		// that's the only int we have
		return 0;		
	}
	if (ad == 0x01) {		// this would be a peripheral card interrupt
		return 0;
	}

	// no other hardware devices at this time, check keyboard/joysticks

	// keyboard reads as an array. Bits 24, 26 and 28 set the line to	
	// scan (columns). Bits 6-14 are used for return. 0 means on.		
	// The address was divided by 2 before being given to the routine	

	// Some hacks here for 99/4 scanning
	if ((ad>=0x03)&&(ad<=0x0a))
	{	
		col=(CRU[0x14]==0 ? 1 : 0) | (CRU[0x13]==0 ? 2 : 0) | (CRU[0x12]==0 ? 4 : 0);	// get column

		if (keyboard==KEY_994A_PS2) {
			// for 99/4A only, not 99/4
			unsigned char in;

			in=CheckTIPolling(col|((CRU[0x15]==0)?8:0));	// add in bit 4 for alpha lock scanning

			if (0xff != in) {
				// (ad-3) is the row number we are checking (bit #)
				if (0 == (in & (1<<(ad-3)))) {
					ret=0;
				} else {
					ret=1;
				}
				return ret;
			}

			// else, try joysticks
			return CheckJoysticks(ad, col);
		}

		// not PS/2, use the old method
		if ((ad==0x07)&&(CRU[0x15]==0))					// is it ALPHA LOCK?
		{	
			ret=0;
			if (GetKeyState(VK_CAPITAL) & 0x01)			// check CAPS LOCK (on?)
			{	
				ret=1;									// set Alpha Lock off (invert caps lock)
			}

			return ret;
		}

		// Either joysticks or keyboard - try joysticks first
		ret = CheckJoysticks(ad, col);
		if (1 == ret) {
			// if nothing else matched, try the keyboard array
			if (key[KEYS[keyboard][col][ad-3]])				// normal key
			{	
					ret=0;
			}
		}
	}

	return(ret);
}

/////////////////////////////////////////////////////////////////////////
// Write a line to the debug buffer displayed on the debug screen
/////////////////////////////////////////////////////////////////////////
void debug_write(char *s, ...)
{
	char buf[1024];

	vsprintf(buf, s, (char*)((&s)+1));

	if (!quitflag) {
		OutputDebugString(buf);
		OutputDebugString("\n");
	}

	buf[DEBUGLEN-1]='\0';

	memcpy(&lines[0][0], &lines[1][0], 33*DEBUGLEN);				// scroll data
	strncpy(&lines[33][0], buf, DEBUGLEN);							// copy in new line
	memset(&lines[33][strlen(buf)], 0x20, DEBUGLEN-strlen(buf));	// clear rest of line
	lines[33][DEBUGLEN-1]='\0';										// zero terminate
	bDebugDirty=true;												// flag redraw
}

#ifdef ALLOW_OLD_HACKS
///////////////////////////////////////////////////////////////////
// Console interrupt simulator
///////////////////////////////////////////////////////////////////
void ConsoleInterrupt()
{
	Byte x1;											// temp var
	int t1,t2;											// temp vars

	interrupt_needed=0;
	x1=rcpubyte(0x83c2);
	if ((x1&0x80)==0)									// if console routine on
	{ 
		if ((x1&0x40)==0)								// update sprites
	    {												// tables are fixed in memory
			for (num_sprites=0; num_sprites<rcpubyte(0x837a); num_sprites++) // note: not masking the sprites!
			{	
				offset=num_sprites<<2;					// I didn't fully understand the sprite
				y_speed=(Byte)VDP[0x780+offset];		// movement routine, and my own simplified
				x_speed=(Byte)VDP[0x781+offset];		// version didn't work as expected. So this
				t1=(Byte)VDP[0x782+offset];				// code is more or less a straight translation
				t2=(Byte)VDP[0x783+offset];				// of the 9900 code ;)
				t1a=(y_speed&0xf0)>>4;
				t1b=(y_speed&0xf);
				t2a=(x_speed&0xf0)>>4;
				t2b=(x_speed&0xf);

				if (t1a>0x7)
				{	
					t1a-=0xf;
					if (t1b==0)
						t1a-=1;
				}
				if (t2a>0x7)
				{	
					t2a-=0xf;
					if (t2b==0)
						t2a-=1;
				}
	
				t1b=(y_speed&0xf);
				if ((y_speed>0x7f)&&(t1b))
					t1b=(t1b-0x10);
				if ((x_speed>0x7f)&&(t2b))
					t2b=(t2b-0x10);

				sprite_y=(Byte)VDP[0x300+offset];
				sprite_x=(Byte)VDP[0x301+offset];
				
				ty=sprite_y;
				tx=sprite_x;
			
				t1+=t1b;
				t2+=t2b;

				sprite_y+=t1a;
				sprite_x+=t2a;
	
				if (t1>0xf)
				{	
					t1-=0xf;
					sprite_y++;
				}
				if (t1<0)
				{	
					t1+=0xf;
					sprite_y--;
				}

				if (t2>0xf)
				{	
					t2-=0xf;
					sprite_x++;
				}
				if (t2<0)
				{	
					t2+=0xf;
					sprite_x--;
				}

				sprite_x&=0xff;
				sprite_y&=0xff;

				if ((sprite_y>=0xc0)&&(sprite_y<=0xe0))
				{	
					if (y_speed>0x7f)
						sprite_y-=0x20;
					else
						sprite_y+=0x20;
				}
	
				VDP[0x300+offset]=(Byte)sprite_y;
				VDP[0x301+offset]=(Byte)sprite_x;
				VDP[0x782+offset]=(Byte)t1;
				VDP[0x783+offset]=(Byte)t2;
			
				if ((ty!=sprite_y)||(tx!=sprite_x))
					redraw_needed=1;

			}
		}
    
		if ((x1&0x20)==0)								// update sound list
	    { 
	      if (rcpubyte(0x83ce))
		  {		wcpubyte(0x83ce, rcpubyte(0x83ce)-1);
				if (rcpubyte(0x83ce)==0)
				{	
					snd_address=romword(0x83cc);
					if (rcpubyte(0x83fd) & 0x01) {
						sndlist=(&VDP[snd_address]);	// VDP sound list
					} else
					{
						sndlist=(&GROMBase[0].GROM[snd_address]);	// GROM sound list
					}
					data=*sndlist++;
					snd_address+=data;
					if ((data!=0)&&(data!=0xff))
					{	
						while (data--)
						{	
							wsndbyte(*sndlist++);
						}
						data=*sndlist++;
						snd_address+=2;
						wrword(0x83cc, snd_address);
						wcpubyte(0x83ce, data);
					}	
					else
					{	
						if (data==0xff)
						{	
							if (rcpubyte(0x83fd) & 0x1)
								wcpubyte(0x83fd, rcpubyte(0x83fd)&0xfe);
							else
								wcpubyte(0x83fd, rcpubyte(0x83fd)|0x01);
						}
						wcpubyte(0x83cc, *sndlist++);
						wcpubyte(0x83cd, *sndlist++);
						wcpubyte(0x83ce, 0x01);
					}
				}
			}
		}
	
	    if ((x1&0x10)==0)								// Check QUIT key
		{ 
			if (((key[VK_MENU])||(key[VK_CONTROL]))&&(key[187]))
			{ 
				while (key['=']);						// wait for key release
				ST=(ST&0xfff0);							// disable interrupts
				WP=romword(0x0000);						// Reset = BLWP @>0000
				wrword(WP+26,0xffff);
				wrword(WP+28,PC);
				wrword(WP+30,ST);
				PC=romword(0x0002);
				for (idx=0; idx<=PCODEGROMBASE; idx++) {
					GROMBase[idx].grmaccess=2;	// no GROM accesses yet
				}
				vdpaccess=0;		// No VDP address writes yet 
				interrupt_needed=0;	// No interrupt missed yet
				nCurrentDSR=-1;		// no DSR selected
				memset(nDSRBank, 0, sizeof(nDSRBank));	// base DSR bank
			}
	    }
	}

	// the following parts can't be turned off, except by LIMI or disabling VDP ints 
  
  	wrword(0x83d6,romword(0x83d6)+2);	// this blanks the screen when it 
										// wraps around to zero 
	
	if (romword(0x83d6)==0)				// we're wrapped					
	{
		t1=rcpubyte(0x83d4);			// read the byte...					
		t1=(t1|0x20)&0xbf;				// int turns ON VDP interrupts, and	
										// turns OFF display bit			
		VDPREG[1]=(Byte)t1;				// write it							
	}
  
	wcpubyte(0x8379, rcpubyte(0x8379)+rcpubyte(0x83fc));// VDP Int Timer - not sure how it	
										// gets this, but here it is. Add	
										// byte from R14 of the GPL WP at
										// >83e0. Can't trace what it is.	
										// E/A manual calls it System Status

	wcpubyte(0x837b, VDPS);				// VDP Status register - copied to CPU RAM	
	VDPS&=0x1f;							// A read clears the register
    
	if (romword(0x83C4)) 				// call the user interrupt routine 
	{	// pretend that we really were called with BLWP by saving the values	
		// into the appropriate workspace (0x83c0)
		
		wrword(0x83da,WP);				// WP in R13 
		wrword(0x83dc,PC);				// PC in R14 
		wrword(0x83de,ST);				// ST in R15 

		ST=(ST&0xfff0);					// disable interrupts

		/* now load the correct workspace, and perform a branch and link to the address */
		WP=0x83e0;				/* interrupt workspace is 0x80e0						*/
		wrword(0x83f6,0x0ab8);	/* write into new R11 hard-coded return address in ROM	*/
		PC=romword(0x83c4);		/* off we go.. TI should return normally now			*/
	}
}
#endif

// Simple thread that watches the event and clears the buffer, then
// restarts playback
void __cdecl SpeechBufThread(void *) {
	DWORD ret; 
	UCHAR *ptr1, *ptr2;
	unsigned long len1, len2;

	for (;;) {
		if (INVALID_HANDLE_VALUE == hSpeechBufferClearEvent) {
			Sleep(150);
			ret=WAIT_TIMEOUT;
		} else {
			ret=WaitForSingleObject(hSpeechBufferClearEvent, 150);
		}
		if (WAIT_OBJECT_0 == ret) {
			if (SUCCEEDED(speechbuf->Lock(0, 0, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER))) {
				// since we haven't started the sound, hopefully the second pointer is nil
				if (len2 != 0) {
					debug_write("Failed to lock speech buffer");
				}
				// signed 16-bit - zero the buffer
				memset(ptr1, 0, len1);

				speechbuf->Unlock(ptr1, len1, ptr2, len2);
			}

			if (FAILED(speechbuf->Play(0, 0, 0))) {
				debug_write("Speech DID NOT START");
			}
		}
		if (quitflag) break;
	}
}

//////////////////////////////////////////////////////////////
// 'Retrace' counter for timing - runs at 50 or 60 hz
// Uses 'hzRate'. Coded for lessor CPU usage, may be
// less accurate on a small scale but about the same
// over time.
//////////////////////////////////////////////////////////////
void __cdecl TimerThread(void *)
{
	MY_LARGE_INTEGER nStart, nEnd, nFreq, nAccum;
	static unsigned long old_total_cycles=0;
	static int oldSystemThrottle=0, oldCPUThrottle=0;
	int nVDPFrames;
	bool bDrawDebug=false;

	// Ensure the scheduler won't move us around on multicore machines
	SetThreadAffinityMask(GetCurrentThread(), 0x01);

	time(&STARTTIME);
	if (FALSE == QueryPerformanceCounter((LARGE_INTEGER*)&nStart)) {
		debug_write("Failed to query performance counter, error 0x%08x\n", GetLastError());
		MessageBox(myWnd, "Unable to run timer system.", "Classic99 Error", MB_ICONSTOP|MB_OK);
		ExitProcess(-1);
	}

	nAccum.QuadPart=0;

	while (quitflag==0) {
		// Check if the system speed has changed
		// This is actually kind of lame - we should use a message or make the vars global
		if ((CPUThrottle != oldCPUThrottle) || (SystemThrottle != oldSystemThrottle)) {
			oldCPUThrottle=CPUThrottle;
			oldSystemThrottle=SystemThrottle;
			old_total_cycles=total_cycles;
			nAccum.QuadPart=0;
		}

		if ((PauseInactive)&&(myWnd != GetForegroundWindow())&&(dbgWnd != GetForegroundWindow())) {
			// Reduce CPU usage when inactive (hack)
			Sleep(100);
		} else {
			// if hzRate==50, then it's 20000us per frame
			// if hzRate==60, then it's 16666us per frame
			switch (CPUThrottle) {
				default:
					Sleep(1000/(hzRate*2));
					break;
				case CPU_OVERDRIVE:
					Sleep(1);	// minimal sleep for load's sake
					break;
				case CPU_MAXIMUM:
					// We do the exchange here since the loop below may not run
					InterlockedExchange((LONG*)&cycles_left, max_cpf*100);
					break;
			}
//			if (SystemThrottle == VDP_CPUSYNC) {
//				nVDPFrames=1;
//			} else {
				nVDPFrames=0;
//			}
			if (FALSE == QueryPerformanceCounter((LARGE_INTEGER*)&nEnd)) {
				debug_write("Failed to query performance counter, error 0x%08x\n", GetLastError());
				MessageBox(myWnd, "Unable to run timer system.", "Classic99 Error", MB_ICONSTOP|MB_OK);
				ExitProcess(-1);
			}
			if (nEnd.QuadPart<nStart.QuadPart) {
				// We wrapped around. This should be a once in a lifetime event, so rather
				// than go nuts, just skip this frame
				nStart.QuadPart=nEnd.QuadPart;
				continue;
			}
			QueryPerformanceFrequency((LARGE_INTEGER*)&nFreq);

			// Work out how long we actually ran for, in microseconds
			nAccum.QuadPart+=(((nEnd.QuadPart-nStart.QuadPart)*1000000i64)/nFreq.QuadPart);	
			nStart.QuadPart=nEnd.QuadPart;					// don't lose any time
			
			nFreq.QuadPart=hzRate==60?16666i64:20000i64;
			while (nAccum.QuadPart > nFreq.QuadPart) {
				nVDPFrames++;
				// this makes us run the right number of frames, and should account for fractions better
				nAccum.QuadPart-=nFreq.QuadPart;

				if (max_cpf > 0) {
					if (CPUThrottle==CPU_NORMAL) {	// don't increment cycles_left if running at infinite speed or paused
						InterlockedExchangeAdd((LONG*)&cycles_left, max_cpf);
					} else {
						InterlockedExchange((LONG*)&cycles_left, max_cpf*50);
					}
				}
			}
			SetEvent(hWakeupEvent);		// wake up CPU if it's sleeping

			if (total_cycles_looped) {
				total_cycles_looped=false;
				old_total_cycles=0;		// mistiming, but survives the wrap.
				// very very fast machines may someday break this loop
			}

			// copy over the speech buffer -- only need to do this once
			// if our timing is right this should always work out about right
			SpeechBufferCopy();

			// TODO: DEFAULT_CPF is based on 60hz
			// This set the VDP processing rate. If VDP overdrive is active,
			// then we base it on the CPU cycles. If not, then we base it on
			// real time.
			if (SystemThrottle == VDP_CPUSYNC) {
				while (old_total_cycles+DEFAULT_CPF <= total_cycles) {
					Counting();					// update counters & VDP interrupt
					old_total_cycles+=DEFAULT_CPF;
					bDrawDebug=true;
				}
			} else {
				while (nVDPFrames > 0) {
					Counting();					// update counters & VDP interrupt
					nVDPFrames--;
					bDrawDebug=true;
				}
			}

			if ((bDrawDebug)&&(dbgWnd)) {
				if (max_cpf > 0) {
					draw_debug();
				}
			}
		}	
	}

	time(&ENDTIME);
	debug_write("Seconds: %ld, ticks: %ld", (long)ENDTIME-STARTTIME, ticks);

	debug_write("Ending Timer Thread");
}

////////////////////////////////////////////////////////////////
// Timer calls this function each tick
////////////////////////////////////////////////////////////////
void Counting()
{
	ticks++;
	retrace_count++;
	if (CRU[2]) {			// VDP interrupts enabled?
		interrupt_needed=1;
	}
}

////////////////////////////////////////////////////////////////
// stpcpy - strcpy then return the address of the \0
////////////////////////////////////////////////////////////////
char *stpcpy(char *dest, char *src)
{
	while (*src) {
		*dest++ = *src++;
	}
	*dest++ = *src++;
  return (dest-1);
}

////////////////////////////////////////////////////////////////
// stpncpy - strncpy then return the address of the \0
////////////////////////////////////////////////////////////////
char *stpncpy(char *dest, char *src, int n)
{
	while (n--) {
		*dest++=*src;
		if (!(*src)) {
			break;
		}
		src++;
	}
	return (dest-1);
}

// Debug step helpers
void DoPause() {
	if (0 != max_cpf) {
		TriggerBreakPoint();
	}
}

void DoStep() {
	if (0 == max_cpf) {
		InterlockedExchange((LONG*)&cycles_left, 1);	// allow one instruction through
		bDebugAfterStep=true;
		SetEvent(hWakeupEvent);		// wake up CPU if it's sleeping
	}
}

void DoPlay() {
	if (0 == max_cpf) {
		max_cpf=oldmax;
		SetWindowText(myWnd, "Classic99");
		InterlockedExchange((LONG*)&cycles_left, max_cpf);
	}
	PostMessage(myWnd, WM_COMMAND, ID_CPUTHROTTLING_NORMAL, 1);
}

void DoFastForward() {
	PostMessage(myWnd, WM_COMMAND, ID_CPUTHROTTLING_SYSTEMMAXIMUM, 1);
}

void TriggerBreakPoint() {
	SetWindowText(myWnd, "Classic99 - Breakpoint. F1 to Continue, F2 to Step");
	max_cpf=0;
	InterlockedExchange(&cycles_left, 0);
	draw_debug();
}
